from pathlib import Path
from datetime import timedelta
import pytest

from kloppy import metrica
from kloppy.domain import (
    Orientation,
    Provider,
    SetPieceType,
    BodyPart,
    EventDataset,
    Point,
    SetPieceQualifier,
    BodyPartQualifier,
)
from kloppy.domain.models.common import DatasetType


class TestMetricaEvents:
    """Tests related to deserialization of Metrica JSON events."""

    @pytest.fixture(scope="class")
    def dataset(self, base_dir: Path) -> EventDataset:
        """Load a Metrica event dataset."""
        dataset = metrica.load_event(
            # FIXME: these files represent different matches
            event_data=base_dir / "files" / "metrica_events.json",
            meta_data=base_dir / "files" / "epts_metrica_metadata.xml",
        )
        assert dataset.dataset_type == DatasetType.EVENT
        assert len(dataset.events) == 3594
        return dataset

    def test_metadata(self, dataset: EventDataset):
        """It should parse the metadata correctly."""
        assert dataset.metadata.provider == Provider.METRICA
        assert len(dataset.metadata.periods) == 2
        assert dataset.metadata.orientation is Orientation.HOME_AWAY
        assert dataset.metadata.teams[0].name == "Team A"
        assert dataset.metadata.teams[1].name == "Team B"
        player = dataset.metadata.teams[0].players[10]
        assert player.player_id == "Track_11"
        assert player.jersey_no == 11
        assert str(player) == "Track_11"
        assert player.starting_position.name == "Goalkeeper"

        assert dataset.metadata.periods[0].id == 1
        assert dataset.metadata.periods[0].start_timestamp == timedelta(
            seconds=18
        )
        assert dataset.metadata.periods[0].end_timestamp == timedelta(
            seconds=19.96
        )
        assert dataset.metadata.periods[1].id == 2
        assert dataset.metadata.periods[1].start_timestamp == timedelta(
            seconds=26
        )
        assert dataset.metadata.periods[1].end_timestamp == timedelta(
            seconds=27.96
        )

    def test_timestamps(self, dataset: EventDataset):
        """It should parse the timestamps correctly."""
        # note: these timestamps are odd because the metadata and event data
        # are from different matches
        assert dataset.events[0].timestamp == timedelta(
            seconds=14.44
        ) - timedelta(
            seconds=450 / 25
        )  # kickoff first half
        assert dataset.events[1749].timestamp == timedelta(
            seconds=2803.6
        ) - timedelta(
            seconds=650 / 25
        )  # kickoff second half

    def test_coordinates(self, dataset: EventDataset):
        """It should parse the coordinates of events correctly."""
        assert dataset.events[0].coordinates == Point(x=0.50125, y=0.48725)

    def test_body_part_qualifiers(self, dataset: EventDataset):
        """It should add body part qualifiers to the event."""
        # The body part qualifier should be set for headers
        header = dataset.get_event_by_id("99")
        assert header.get_qualifier_value(BodyPartQualifier) == BodyPart.HEAD
        # It should be None (i.e., unknown) for events that are not headers
        foot = dataset.get_event_by_id("2")
        assert foot.get_qualifier_value(BodyPartQualifier) is None

    def test_set_piece(self, dataset: EventDataset):
        """It should integrate set piece events in the next event."""
        # The next event can be a pass
        kick_off_event = dataset.get_event_by_id("1")
        assert kick_off_event is None
        pass_after_kick_off = dataset.get_event_by_id("2")
        assert (
            pass_after_kick_off.get_qualifier_value(SetPieceQualifier)
            == SetPieceType.KICK_OFF
        )
        # or a shot
        free_kick_event = dataset.get_event_by_id("130")
        assert free_kick_event is None
        shot_after_free_kick = dataset.get_event_by_id("131")
        assert (
            shot_after_free_kick.get_qualifier_value(SetPieceQualifier)
            == SetPieceType.FREE_KICK
        )



"""Convert StatsBomb event stream data to a kloppy EventDataset."""

from .deserializer import StatsBombDeserializer, StatsBombInputs

__all__ = [
    "StatsBombDeserializer",
    "StatsBombInputs",
]



from kloppy.domain import FormationType, PositionType

formation_id_mapping = {
    2: FormationType.FOUR_FOUR_TWO,
    3: FormationType.FOUR_ONE_TWO_ONE_TWO,
    4: FormationType.FOUR_THREE_THREE,
    5: FormationType.FOUR_FIVE_ONE,
    6: FormationType.FOUR_FOUR_ONE_ONE,
    7: FormationType.FOUR_ONE_FOUR_ONE,
    8: FormationType.FOUR_TWO_THREE_ONE,
    9: FormationType.FOUR_THREE_TWO_ONE,
    10: FormationType.FIVE_THREE_TWO,
    11: FormationType.FIVE_FOUR_ONE,
    12: FormationType.THREE_FIVE_TWO,
    13: FormationType.THREE_FOUR_THREE,
    14: FormationType.THREE_ONE_THREE_ONE_TWO,
    15: FormationType.FOUR_TWO_TWO_TWO,
    16: FormationType.THREE_FIVE_ONE_ONE,
    17: FormationType.THREE_FOUR_TWO_ONE,
    18: FormationType.THREE_FOUR_ONE_TWO,
    19: FormationType.THREE_ONE_FOUR_TWO,
    20: FormationType.THREE_ONE_TWO_ONE_THREE,
    21: FormationType.FOUR_ONE_THREE_TWO,
    22: FormationType.FOUR_TWO_FOUR_ZERO,
    23: FormationType.FOUR_THREE_ONE_TWO,
    24: FormationType.THREE_TWO_FOUR_ONE,
    25: FormationType.THREE_THREE_THREE_ONE,
}

formation_name_mapping = {
    "442": FormationType.FOUR_FOUR_TWO,
    "41212": FormationType.FOUR_ONE_TWO_ONE_TWO,
    "433": FormationType.FOUR_THREE_THREE,
    "451": FormationType.FOUR_FIVE_ONE,
    "4411": FormationType.FOUR_FOUR_ONE_ONE,
    "4141": FormationType.FOUR_ONE_FOUR_ONE,
    "4231": FormationType.FOUR_TWO_THREE_ONE,
    "4321": FormationType.FOUR_THREE_TWO_ONE,
    "532": FormationType.FIVE_THREE_TWO,
    "541": FormationType.FIVE_FOUR_ONE,
    "352": FormationType.THREE_FIVE_TWO,
    "343": FormationType.THREE_FOUR_THREE,
    "4222": FormationType.FOUR_TWO_TWO_TWO,
    "3511": FormationType.THREE_FIVE_ONE_ONE,
    "3421": FormationType.THREE_FOUR_TWO_ONE,
    "3412": FormationType.THREE_FOUR_ONE_TWO,
    "3142": FormationType.THREE_ONE_FOUR_TWO,
    "343d": FormationType.THREE_ONE_TWO_ONE_THREE,
    "4132": FormationType.FOUR_ONE_THREE_TWO,
    "4312": FormationType.FOUR_THREE_ONE_TWO,
    "3241": FormationType.THREE_TWO_FOUR_ONE,
    "3331": FormationType.THREE_THREE_THREE_ONE,
}

formation_position_mapping = {
    FormationType.FOUR_FOUR_TWO: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightBack,
        3: PositionType.LeftBack,
        4: PositionType.RightCentralMidfield,
        5: PositionType.RightCenterBack,
        6: PositionType.LeftCenterBack,
        7: PositionType.RightMidfield,
        8: PositionType.LeftCentralMidfield,
        9: PositionType.LeftForward,
        10: PositionType.RightForward,
        11: PositionType.LeftMidfield,
    },
    FormationType.FOUR_ONE_TWO_ONE_TWO: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightBack,
        3: PositionType.LeftBack,
        4: PositionType.CenterDefensiveMidfield,
        5: PositionType.RightCenterBack,
        6: PositionType.LeftCenterBack,
        7: PositionType.RightMidfield,
        8: PositionType.CenterAttackingMidfield,
        9: PositionType.LeftForward,
        10: PositionType.RightForward,
        11: PositionType.LeftMidfield,
    },
    FormationType.FOUR_THREE_THREE: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightBack,
        3: PositionType.LeftBack,
        4: PositionType.CenterMidfield,
        5: PositionType.RightCenterBack,
        6: PositionType.LeftCenterBack,
        7: PositionType.RightMidfield,
        8: PositionType.LeftMidfield,
        9: PositionType.Striker,
        10: PositionType.RightWing,
        11: PositionType.LeftWing,
    },
    FormationType.FOUR_FIVE_ONE: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightBack,
        3: PositionType.LeftBack,
        4: PositionType.RightCentralMidfield,
        5: PositionType.RightCenterBack,
        6: PositionType.LeftCenterBack,
        7: PositionType.RightMidfield,
        8: PositionType.LeftCentralMidfield,
        9: PositionType.Striker,
        10: PositionType.CenterMidfield,
        11: PositionType.LeftMidfield,
    },
    FormationType.FOUR_FOUR_ONE_ONE: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightBack,
        3: PositionType.LeftBack,
        4: PositionType.RightCentralMidfield,
        5: PositionType.RightCenterBack,
        6: PositionType.LeftCenterBack,
        7: PositionType.RightMidfield,
        8: PositionType.LeftCentralMidfield,
        9: PositionType.Striker,
        10: PositionType.CenterAttackingMidfield,
        11: PositionType.LeftMidfield,
    },
    FormationType.FOUR_ONE_FOUR_ONE: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightBack,
        3: PositionType.LeftBack,
        4: PositionType.CenterDefensiveMidfield,
        5: PositionType.RightCenterBack,
        6: PositionType.LeftCenterBack,
        7: PositionType.RightMidfield,
        8: PositionType.RightCentralMidfield,
        9: PositionType.Striker,
        10: PositionType.LeftCentralMidfield,
        11: PositionType.LeftMidfield,
    },
    FormationType.FOUR_TWO_THREE_ONE: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightBack,
        3: PositionType.LeftBack,
        4: PositionType.RightCentralMidfield,
        5: PositionType.RightCenterBack,
        6: PositionType.LeftCenterBack,
        7: PositionType.RightAttackingMidfield,
        8: PositionType.LeftCentralMidfield,
        9: PositionType.Striker,
        10: PositionType.CenterAttackingMidfield,
        11: PositionType.LeftAttackingMidfield,
    },
    FormationType.FOUR_THREE_TWO_ONE: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightBack,
        3: PositionType.LeftBack,
        4: PositionType.CenterMidfield,
        5: PositionType.RightCenterBack,
        6: PositionType.LeftCenterBack,
        7: PositionType.LeftMidfield,
        8: PositionType.RightMidfield,
        9: PositionType.Striker,
        10: PositionType.RightAttackingMidfield,
        11: PositionType.LeftAttackingMidfield,
    },
    FormationType.FIVE_THREE_TWO: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightBack,
        3: PositionType.LeftBack,
        4: PositionType.LeftCenterBack,
        5: PositionType.CenterBack,
        6: PositionType.RightCenterBack,
        7: PositionType.RightCentralMidfield,
        8: PositionType.CenterMidfield,
        9: PositionType.LeftForward,
        10: PositionType.RightForward,
        11: PositionType.LeftCentralMidfield,
    },
    FormationType.FIVE_FOUR_ONE: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightBack,
        3: PositionType.LeftBack,
        4: PositionType.LeftCenterBack,
        5: PositionType.CenterBack,
        6: PositionType.RightCenterBack,
        7: PositionType.RightMidfield,
        8: PositionType.RightCentralMidfield,
        9: PositionType.Striker,
        10: PositionType.LeftCentralMidfield,
        11: PositionType.LeftMidfield,
    },
    FormationType.THREE_FIVE_TWO: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightMidfield,
        3: PositionType.LeftMidfield,
        4: PositionType.LeftCenterBack,
        5: PositionType.CenterBack,
        6: PositionType.RightCenterBack,
        7: PositionType.RightCentralMidfield,
        8: PositionType.LeftCentralMidfield,
        9: PositionType.LeftForward,
        10: PositionType.RightForward,
        11: PositionType.CenterMidfield,
    },
    FormationType.THREE_FOUR_THREE: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightMidfield,
        3: PositionType.LeftMidfield,
        4: PositionType.LeftCenterBack,
        5: PositionType.CenterBack,
        6: PositionType.RightCenterBack,
        7: PositionType.RightCentralMidfield,
        8: PositionType.LeftCentralMidfield,
        9: PositionType.Striker,
        10: PositionType.RightWing,
        11: PositionType.LeftWing,
    },
    FormationType.THREE_ONE_THREE_ONE_TWO: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.Unknown,
        3: PositionType.Unknown,
        4: PositionType.Unknown,
        5: PositionType.Unknown,
        6: PositionType.Unknown,
        7: PositionType.Unknown,
        8: PositionType.Unknown,
        9: PositionType.Unknown,
        10: PositionType.Unknown,
        11: PositionType.Unknown,
    },
    FormationType.FOUR_TWO_TWO_TWO: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightBack,
        3: PositionType.LeftBack,
        4: PositionType.RightDefensiveMidfield,
        5: PositionType.RightCenterBack,
        6: PositionType.LeftCenterBack,
        7: PositionType.RightCentralMidfield,
        8: PositionType.LeftDefensiveMidfield,
        9: PositionType.LeftForward,
        10: PositionType.RightForward,
        11: PositionType.LeftCentralMidfield,
    },
    FormationType.THREE_FIVE_ONE_ONE: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightMidfield,
        3: PositionType.LeftMidfield,
        4: PositionType.LeftCenterBack,
        5: PositionType.CenterBack,
        6: PositionType.RightCenterBack,
        7: PositionType.RightDefensiveMidfield,
        8: PositionType.LeftDefensiveMidfield,
        9: PositionType.Striker,
        10: PositionType.CenterAttackingMidfield,
        11: PositionType.CenterDefensiveMidfield,
    },
    FormationType.THREE_FOUR_TWO_ONE: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightMidfield,
        3: PositionType.LeftMidfield,
        4: PositionType.LeftCenterBack,
        5: PositionType.CenterBack,
        6: PositionType.RightCenterBack,
        7: PositionType.RightCentralMidfield,
        8: PositionType.LeftCentralMidfield,
        9: PositionType.Striker,
        10: PositionType.RightForward,
        11: PositionType.LeftForward,
    },
    FormationType.THREE_FOUR_ONE_TWO: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightMidfield,
        3: PositionType.LeftMidfield,
        4: PositionType.LeftCenterBack,
        5: PositionType.CenterBack,
        6: PositionType.RightCenterBack,
        7: PositionType.RightCentralMidfield,
        8: PositionType.LeftCentralMidfield,
        9: PositionType.CenterAttackingMidfield,
        10: PositionType.RightForward,
        11: PositionType.LeftForward,
    },
    FormationType.THREE_ONE_FOUR_TWO: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightMidfield,
        3: PositionType.LeftMidfield,
        4: PositionType.CenterBack,
        5: PositionType.RightCenterBack,
        6: PositionType.LeftCenterBack,
        7: PositionType.RightCentralMidfield,
        8: PositionType.CenterDefensiveMidfield,
        9: PositionType.RightForward,
        10: PositionType.LeftWing,
        11: PositionType.LeftCentralMidfield,
    },
    FormationType.THREE_ONE_TWO_ONE_THREE: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightMidfield,
        3: PositionType.LeftMidfield,
        4: PositionType.LeftCenterBack,
        5: PositionType.CenterBack,
        6: PositionType.RightCenterBack,
        7: PositionType.CenterAttackingMidfield,
        8: PositionType.CenterDefensiveMidfield,
        9: PositionType.Striker,
        10: PositionType.RightWing,
        11: PositionType.LeftWing,
    },
    FormationType.FOUR_ONE_THREE_TWO: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightBack,
        3: PositionType.LeftBack,
        4: PositionType.CenterDefensiveMidfield,
        5: PositionType.RightCenterBack,
        6: PositionType.LeftCenterBack,
        7: PositionType.RightWing,
        8: PositionType.CenterAttackingMidfield,
        9: PositionType.RightForward,
        10: PositionType.LeftForward,
        11: PositionType.LeftWing,
    },
    FormationType.FOUR_TWO_FOUR_ZERO: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightBack,
        3: PositionType.LeftBack,
        4: PositionType.RightCentralMidfield,
        5: PositionType.RightCenterBack,
        6: PositionType.LeftCenterBack,
        7: PositionType.RightWing,
        8: PositionType.LeftCentralMidfield,
        9: PositionType.RightForward,
        10: PositionType.LeftForward,
        11: PositionType.LeftWing,
    },
    FormationType.FOUR_THREE_ONE_TWO: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightBack,
        3: PositionType.LeftBack,
        4: PositionType.CenterMidfield,
        5: PositionType.RightCenterBack,
        6: PositionType.LeftCenterBack,
        7: PositionType.RightMidfield,
        8: PositionType.CenterAttackingMidfield,
        9: PositionType.RightForward,
        10: PositionType.LeftForward,
        11: PositionType.LeftMidfield,
    },
    FormationType.THREE_TWO_FOUR_ONE: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightDefensiveMidfield,
        3: PositionType.LeftDefensiveMidfield,
        4: PositionType.LeftCenterBack,
        5: PositionType.CenterBack,
        6: PositionType.RightCenterBack,
        7: PositionType.RightCentralMidfield,
        8: PositionType.LeftCentralMidfield,
        9: PositionType.Striker,
        10: PositionType.RightWing,
        11: PositionType.LeftWing,
    },
    FormationType.THREE_THREE_THREE_ONE: {
        0: None,
        1: PositionType.Goalkeeper,
        2: PositionType.RightDefensiveMidfield,
        3: PositionType.LeftDefensiveMidfield,
        4: PositionType.LeftCenterBack,
        5: PositionType.CenterBack,
        6: PositionType.RightCenterBack,
        7: PositionType.CenterAttackingMidfield,
        8: PositionType.CenterDefensiveMidfield,
        9: PositionType.Striker,
        10: PositionType.RightWing,
        11: PositionType.LeftWing,
    },
}



from datetime import timedelta, datetime, timezone
from pathlib import Path
from typing import List

import pytest

from kloppy.domain import (
    Provider,
    Orientation,
    Point,
    Point3D,
    DatasetType,
    Ground,
    BallState,
)

from kloppy import signality


class TestSignalityTracking:
    @pytest.fixture
    def meta_data(self, base_dir) -> Path:
        return base_dir / "files/signality_meta_data.json"

    @pytest.fixture
    def venue_information(self, base_dir) -> Path:
        return base_dir / "files/signality_venue_information.json"

    @pytest.fixture
    def raw_data_feeds(self, base_dir) -> List[Path]:
        return [
            base_dir / "files/signality_p1_raw_data_subset.json",
            base_dir / "files/signality_p2_raw_data_subset.json",
        ]

    def test_correct_deserialization(
        self,
        raw_data_feeds: List[Path],
        meta_data: Path,
        venue_information: Path,
    ):
        dataset = signality.load(
            meta_data=meta_data,
            raw_data_feeds=raw_data_feeds,
            venue_information=venue_information,
            coordinates="signality",
        )

        assert dataset.metadata.provider == Provider.SIGNALITY
        assert dataset.dataset_type == DatasetType.TRACKING
        assert len(dataset.records) == 10
        assert len(dataset.metadata.periods) == 2
        assert dataset.metadata.orientation == Orientation.HOME_AWAY
        assert dataset.metadata.periods[0].id == 1
        assert dataset.metadata.periods[0].start_timestamp == datetime(
            2024, 10, 6, 12, 0, 1, 408000, timezone.utc
        )
        assert dataset.metadata.periods[0].end_timestamp == datetime(
            2024, 10, 6, 12, 48, 43, 858000, timezone.utc
        )
        assert dataset.metadata.periods[1].id == 2
        assert dataset.metadata.periods[1].start_timestamp == datetime(
            2024, 10, 6, 13, 4, 45, 775000, timezone.utc
        )
        assert dataset.metadata.periods[1].end_timestamp == datetime(
            2024, 10, 6, 13, 54, 11, 730000, timezone.utc
        )

        start_frame = dataset.records[0]
        assert start_frame.frame_id == 0
        assert start_frame.timestamp == timedelta(milliseconds=12)

        end_frame = dataset.records[-1]
        assert end_frame.frame_id == 147211
        assert end_frame.timestamp == timedelta(seconds=2965, milliseconds=955)

        home_team = dataset.metadata.teams[0]
        assert home_team.ground == Ground.HOME
        assert len(home_team.players) == 20

        home_team_gk = home_team.get_player_by_id("home_45")
        start_frame_home_team_gk_data = start_frame.players_data[home_team_gk]
        assert start_frame_home_team_gk_data.coordinates == Point(
            x=-47.78, y=0.59
        )
        assert start_frame_home_team_gk_data.speed == 0.65

        away_team = dataset.metadata.teams[1]
        assert away_team.ground == Ground.AWAY
        assert len(away_team.players) == 20

        away_team_gk = away_team.get_player_by_id("away_1")
        start_frame_away_team_gk_data = start_frame.players_data[away_team_gk]
        assert start_frame_away_team_gk_data.coordinates == Point(
            x=46.48, y=0.23
        )
        assert start_frame_away_team_gk_data.speed == 0.942

        assert start_frame.ball_coordinates == Point3D(x=-41.18, y=0.08, z=0)

        first_second_half_frame = next(
            frame for frame in dataset.frames if frame.period.id == 2
        )
        assert first_second_half_frame.frame_id == 73063

        first_second_half_frame_home_team_gk_data = (
            first_second_half_frame.players_data[home_team_gk]
        )
        assert first_second_half_frame_home_team_gk_data.coordinates == Point(
            x=45.37, y=0.22
        )

        pitch_dimensions = dataset.metadata.pitch_dimensions
        assert pitch_dimensions.x_dim.min == -104.82 / 2
        assert pitch_dimensions.x_dim.max == 104.82 / 2
        assert pitch_dimensions.y_dim.min == -68.407 / 2
        assert pitch_dimensions.y_dim.max == 68.407 / 2

        assert start_frame.ball_state == BallState.ALIVE

        assert end_frame.ball_state == BallState.DEAD



import bz2
import gzip
import json
import lzma
import os
import sys
import zipfile
from io import BytesIO
from pathlib import Path

import pytest
from botocore.session import Session
from moto.moto_server.threaded_moto_server import ThreadedMotoServer

from kloppy.config import set_config
from kloppy.exceptions import InputNotFoundError
from kloppy.io import expand_inputs, get_file_extension, open_as_file


@pytest.fixture()
def filesystem_content(tmp_path: Path) -> Path:
    """Set up the content to be read from a local filesystem."""
    content = "Hello, world!"
    content_bytes = content.encode("utf-8")

    # Create a regular text file
    text_file = tmp_path / "testfile.txt"
    text_file.write_text(content)

    # Create a gzip-compressed file
    gz_file = tmp_path / "testfile.txt.gz"
    with gzip.open(gz_file, "wb") as f_out:
        f_out.write(content_bytes)

    # Create a xz-compressed file
    xz_file = tmp_path / "testfile.txt.xz"
    with lzma.open(xz_file, "wb") as f_out:
        f_out.write(content_bytes)

    # Create a bzip2-compressed file
    bz2_file = tmp_path / "testfile.txt.bz2"
    with bz2.open(bz2_file, "wb") as f_out:
        f_out.write(content_bytes)

    return tmp_path


class TestOpenAsFile:
    """Tests for the open_as_file function."""

    def test_bytes(self):
        """It should be able to open a bytes object as a file."""
        with open_as_file(b"Hello, world!") as fp:
            assert fp is not None
            assert fp.read() == b"Hello, world!"

    def test_data_string(self):
        """It should be able to open a json/xml string as a file."""
        with open_as_file('{"msg": "Hello, world!"}') as fp:
            assert fp is not None
            assert json.load(fp) == {"msg": "Hello, world!"}

    def test_stream(self):
        """It should be able to open a byte stream as a file."""
        data = b"Hello, world!"
        with open_as_file(BytesIO(data)) as fp:
            assert fp is not None
            assert fp.read() == data

    @pytest.mark.parametrize(
        "compress_func",
        [
            gzip.compress,
            bz2.compress,
            lzma.compress,
        ],
        ids=["gzip", "bz2", "xz"],
    )
    def test_compressed_stream(self, compress_func):
        """It should be able to open a compressed byte stream as a file."""
        data = compress_func(b"Hello, world!")
        with open_as_file(BytesIO(data)) as fp:
            assert fp is not None
            assert fp.read() == b"Hello, world!"

    def test_path_str(self, filesystem_content: Path):
        """It should be able to open a file from a string path."""
        path = str(filesystem_content / "testfile.txt")
        with open_as_file(path) as fp:
            assert fp is not None
            assert fp.read() == b"Hello, world!"

    def test_path_obj(self, filesystem_content: Path):
        """It should be able to open a file from a Path object."""
        path = filesystem_content / "testfile.txt"
        with open_as_file(path) as fp:
            assert fp is not None
            assert fp.read() == b"Hello, world!"

    @pytest.mark.parametrize("ext", ["gz", "xz", "bz2"])
    def test_path_compressed(self, filesystem_content: Path, ext: str):
        """It should be able to open a compressed local file."""
        path = filesystem_content / f"testfile.txt.{ext}"
        with open_as_file(path) as fp:
            assert fp is not None
            assert fp.read() == b"Hello, world!"

    def test_path_missing(self, filesystem_content: Path):
        """It should raise an error if the file is not found."""
        path = filesystem_content / "missing.txt"
        with pytest.raises(InputNotFoundError):
            with open_as_file(path) as _:
                pass


class TestExpandInputs:
    @pytest.fixture
    def mock_filesystem(self, tmp_path):
        # Create a temporary directory structure
        file1 = tmp_path / "file1.txt"
        file2 = tmp_path / "file2.log"
        subdir = tmp_path / "subdir"
        subdir.mkdir()
        file3 = subdir / "file3.txt"

        file1.write_text("Content of file1")
        file2.write_text("Content of file2")
        file3.write_text("Content of file3")

        return {
            "root": str(tmp_path.as_posix()),
            "file1": str(file1.as_posix()),
            "file2": str(file2.as_posix()),
            "subdir": str(subdir.as_posix()),
            "file3": str(file3.as_posix()),
        }

    def test_single_file(self, mock_filesystem):
        files = list(expand_inputs(mock_filesystem["file1"]))
        assert files == [mock_filesystem["file1"]]

    def test_directory_expansion(self, mock_filesystem):
        files = sorted(expand_inputs(mock_filesystem["root"]))
        expected_files = sorted(
            [
                mock_filesystem["file1"],
                mock_filesystem["file2"],
                mock_filesystem["file3"],
            ]
        )
        assert files == expected_files

    def test_regex_filter(self, mock_filesystem):
        files = list(
            expand_inputs(mock_filesystem["root"], regex_filter=r".*.txt$")
        )
        expected_files = [
            mock_filesystem["file1"],
            mock_filesystem["file3"],
        ]
        assert sorted(files) == sorted(expected_files)

    def test_sort_key(self, mock_filesystem):
        files = list(
            expand_inputs(mock_filesystem["root"], sort_key=lambda x: x[::-1])
        )
        expected_files = sorted(
            [
                mock_filesystem["file1"],
                mock_filesystem["file2"],
                mock_filesystem["file3"],
            ],
            key=lambda x: x[::-1],
        )
        assert files == expected_files

    def test_list_of_files(self, mock_filesystem):
        input_list = [mock_filesystem["file1"], mock_filesystem["file2"]]
        files = list(expand_inputs(input_list))
        assert files == input_list

    def test_invalid_path(self):
        with pytest.raises(InputNotFoundError):
            list(expand_inputs("nonexistent_file.txt"))


def test_get_file_extension():
    assert get_file_extension(Path("data.xml")) == ".xml"
    assert get_file_extension("data.xml") == ".xml"
    assert get_file_extension("data.xml.gz") == ".xml"
    assert get_file_extension("data") == ""


class TestHTTPAdapter:
    @pytest.fixture(autouse=True)
    def httpserver_content(self, httpserver):
        """Set up the content to be read from an HTTP server."""
        # Define the content
        content = "Hello, world!"
        compressed_content = gzip.compress(b"Hello, world!")

        # Serve the plain text file
        httpserver.expect_request("/testfile.txt").respond_with_data(content)

        # Serve the compressed text file with Content-Encoding header
        httpserver.expect_request(
            "/compressed_testfile.txt"
        ).respond_with_data(
            compressed_content,
            headers={"Content-Encoding": "gzip", "Content-Type": "text/plain"},
        )

        # Serve the gzip file with application/x-gzip content type
        httpserver.expect_request("/testfile.txt.gz").respond_with_data(
            compressed_content,
            headers={"Content-Type": "application/x-gzip"},
        )

        # Generate the index.html content with links to all resources
        index_html = f"""
        <html>
            <head><title>Test Content</title></head>
            <body>
                <h1>Available Data</h1>
                <ul>
                    <li><a href="/testfile.txt">Plain Text File</a></li>
                    <li><a href="/compressed_testfile.txt">Compressed Text File (gzip)</a></li>
                    <li><a href="{httpserver.url_for("/testfile.txt.gz")}">Gzip File</a></li>
                </ul>
            </body>
        </html>
        """

        # Serve the index.html page
        httpserver.expect_request("/").respond_with_data(
            index_html, headers={"Content-Type": "text/html"}
        )

        return httpserver

    def test_expand_inputs(self, httpserver):
        """It should be able to list the contents of an HTTP server."""
        url = httpserver.url_for("/")
        assert list(expand_inputs(url)) == [
            httpserver.url_for("/compressed_testfile.txt"),
            httpserver.url_for("/testfile.txt"),
            httpserver.url_for("/testfile.txt.gz"),
        ]

    def test_open_as_file(self, httpserver):
        """It should be able to open a file from a URL."""
        url = httpserver.url_for("/testfile.txt")
        with open_as_file(url) as fp:
            assert fp is not None
            assert fp.read() == b"Hello, world!"

    def test_open_as_file_compressed(self, httpserver):
        """It should be able to open a compressed file from a URL."""
        # If the server returns a content-encoding header, the file should be
        # decompressed by the request library
        url = httpserver.url_for("/compressed_testfile.txt")
        with open_as_file(url) as fp:
            assert fp is not None
            assert fp.read() == b"Hello, world!"

        # If the server does not set a content-type header, but the URL ends
        # with .gz, the file should be decompressed by kloppy
        url = httpserver.url_for("/testfile.txt.gz")
        with open_as_file(url) as fp:
            assert fp is not None
            assert fp.read() == b"Hello, world!"


@pytest.mark.skipif(
    sys.version_info < (3, 9), reason="Patch requires Python 3.9 or higher"
)
class TestS3Adapter:
    endpoint_uri = "http://127.0.0.1:5555"
    test_bucket_name = "test-bucket"
    files = {
        "testfile.txt": b"Hello, world!",
        "testfile.txt.gz": gzip.compress(b"Hello, world!"),
    }

    @pytest.fixture(scope="class", autouse=True)
    def s3_content(self):
        """Set up the content to be read from a S3 bucket."""
        server = ThreadedMotoServer(ip_address="127.0.0.1", port=5555)
        server.start()
        if "AWS_SECRET_ACCESS_KEY" not in os.environ:
            os.environ["AWS_SECRET_ACCESS_KEY"] = "foo"
        if "AWS_ACCESS_KEY_ID" not in os.environ:
            os.environ["AWS_ACCESS_KEY_ID"] = "foo"

        session = Session()
        client = session.create_client(
            "s3", endpoint_url=self.endpoint_uri, region_name="us-east-1"
        )
        client.create_bucket(Bucket=self.test_bucket_name, ACL="public-read")

        for f, data in self.files.items():
            client.put_object(Bucket=self.test_bucket_name, Key=f, Body=data)

        yield

        server.stop()

    @pytest.fixture(scope="class", autouse=True)
    def s3fs(self):
        """Set up the S3FileSystem."""
        from s3fs import S3FileSystem

        s3 = S3FileSystem(
            anon=False, client_kwargs={"endpoint_url": self.endpoint_uri}
        )
        set_config("adapters.s3.s3fs", s3)

    def test_list_directory(self):
        """It should be able to list the contents of an S3 bucket."""
        assert set(expand_inputs("s3://test-bucket/")) == {
            "s3://test-bucket/testfile.txt",
            "s3://test-bucket/testfile.txt.gz",
        }

    def test_open_as_file(self):
        """It should be able to open a file from an S3 bucket."""
        with open_as_file("s3://test-bucket/testfile.txt") as fp:
            assert fp is not None
            assert fp.read() == b"Hello, world!"

    def test_open_as_file_compressed(self):
        """It should be able to open a file from an S3 bucket."""
        with open_as_file("s3://test-bucket/testfile.txt.gz") as fp:
            assert fp is not None
            assert fp.read() == b"Hello, world!"


class TestZipAdapter:
    @pytest.fixture()
    def zip_archive_content(self, tmp_path: Path) -> Path:
        """
        Set up a ZIP archive containing test files.
        """
        zip_path = tmp_path / "archive.zip"

        # Create a text file to include in the ZIP archive
        text_file_path = tmp_path / "testfile.txt"
        with open(text_file_path, "w") as f:
            f.write("Hello, world!")

        # Create a ZIP archive and add the text file to it
        with zipfile.ZipFile(zip_path, "w") as zipf:
            zipf.write(text_file_path, arcname="testfile.txt")

        # Optionally, add more files to the ZIP archive
        another_file_path = tmp_path / "anothertestfile.txt"
        with open(another_file_path, "w") as f:
            f.write("Another file content")

        with zipfile.ZipFile(zip_path, "a") as zipf:
            zipf.write(another_file_path, arcname="anothertestfile.txt")

        set_config("adapters.zip.fo", str(zip_path))
        return zip_path

    def test_list_directory(self, zip_archive_content):
        """It should be able to list the contents of a zip archive."""
        assert list(expand_inputs("zip:///")) == [
            str("zip://anothertestfile.txt"),
            str("zip://testfile.txt"),
        ]

    def test_open_as_file(self, zip_archive_content):
        """It should be able to open a file from a URL."""
        with open_as_file("zip://testfile.txt") as fp:
            assert fp is not None
            assert fp.read() == b"Hello, world!"



from typing import Union, Optional

from .domain import (
    Dataset,
    Orientation,
    PitchDimensions,
    DatasetTransformer,
    Provider,
    build_coordinate_system,
    CoordinateSystem,
)


def transform(
    dataset: Dataset,
    to_orientation: Optional[Union[Orientation, str]] = None,
    to_pitch_dimensions: Optional[PitchDimensions] = None,
    to_coordinate_system: Optional[
        Union[CoordinateSystem, Provider, str]
    ] = None,
) -> Dataset:
    # convert raw orientation to object
    if to_orientation is not None and isinstance(to_orientation, str):
        to_orientation = Orientation[to_orientation.upper()]

    # convert raw coordinate system to object
    if to_coordinate_system is not None:
        if isinstance(to_coordinate_system, str):
            to_coordinate_system = build_coordinate_system(
                provider=Provider[to_coordinate_system.upper()],
                pitch_length=dataset.metadata.coordinate_system.pitch_length,
                pitch_width=dataset.metadata.coordinate_system.pitch_width,
            )
        elif isinstance(to_coordinate_system, Provider):
            to_coordinate_system = build_coordinate_system(
                provider=to_coordinate_system,
                pitch_length=dataset.metadata.coordinate_system.pitch_length,
                pitch_width=dataset.metadata.coordinate_system.pitch_width,
            )

    return DatasetTransformer.transform_dataset(
        dataset=dataset,
        to_orientation=to_orientation,
        to_coordinate_system=to_coordinate_system,
        to_pitch_dimensions=to_pitch_dimensions,
    )



from datetime import datetime, timezone
from typing import IO, List, Optional, Tuple

from lxml import objectify

from kloppy.domain import Ground, Orientation, Period, Team

from .base import TracabMetadataParser
from .common import create_period, create_team


class TracabFlatXMLMetadataParser(TracabMetadataParser):
    """Extract metadata from an XML file with a flat metadata object.

    Expected XML structure:

        <root>
            <GameID>13331</GameID>
            <CompetitionID>55</CompetitionID>
            ...
        </root>

    Args:
        feed : The data stream of a game to parse.

    Attributes:
        root : The root of the XML tree.
    """

    def __init__(self, feed: IO[bytes]) -> None:
        self.root = objectify.fromstring(feed.read())

    def extract_periods(self) -> List[Period]:
        frame_rate = self.extract_frame_rate()
        periods = []
        for i in [1, 2, 3, 4, 5]:
            start_frame_id = int(self.root[f"Phase{i}StartFrame"])
            end_frame_id = int(self.root[f"Phase{i}EndFrame"])
            period = create_period(i, start_frame_id, end_frame_id, frame_rate)
            if period:
                periods.append(period)
        return periods

    def extract_date(self) -> Optional[datetime]:
        date = datetime.strptime(
            str(self.root["Kickoff"]), "%Y-%m-%d %H:%M:%S"
        ).replace(tzinfo=timezone.utc)
        return date

    def extract_game_id(self) -> Optional[str]:
        game_id = str(self.root["GameID"])
        return game_id

    def extract_lineups(self) -> Tuple[Team, Team]:
        start_frame_id = int(self.root["Phase1StartFrame"])
        if hasattr(self.root, "HomeTeam") and hasattr(self.root, "AwayTeam"):
            home_team = create_team(
                self.root["HomeTeam"],
                Ground.HOME,
                start_frame_id,
                id_suffix="ID",
                player_item="item",
            )
            away_team = create_team(
                self.root["AwayTeam"],
                Ground.AWAY,
                start_frame_id,
                id_suffix="ID",
                player_item="item",
            )
        else:
            home_team = Team(team_id="home", name="home", ground=Ground.HOME)
            away_team = Team(team_id="away", name="away", ground=Ground.AWAY)
        return (home_team, away_team)

    def extract_pitch_dimensions(self) -> Tuple[float, float]:
        pitch_size_width = float(self.root["PitchShortSide"]) / 100
        pitch_size_length = float(self.root["PitchLongSide"]) / 100
        return pitch_size_length, pitch_size_width

    def extract_frame_rate(self) -> int:
        frame_rate = int(self.root["FrameRate"])
        return frame_rate

    def extract_orientation(self) -> Orientation:
        if self.root.get("Phase1HomeGKLeft", None) is not None:
            orientation = (
                Orientation.HOME_AWAY
                if bool(self.root["Phase1HomeGKLeft"])
                else Orientation.AWAY_HOME
            )
            return orientation
        else:
            return None



from typing import BinaryIO, List

import fsspec

from kloppy.config import get_config
from kloppy.exceptions import AdapterError, InputNotFoundError

from .adapter import Adapter
from .fsspec import FSSpecAdapter

try:
    from js import XMLHttpRequest  # noqa: F401

    RUNS_IN_BROWSER = True
except ImportError:
    RUNS_IN_BROWSER = False


def check_requests_patch():
    if RUNS_IN_BROWSER:
        try:
            import pyodide_http
        except ImportError:
            raise AdapterError(
                "Seems like you don't have `pyodide-http` installed, which is required to make http requests "
                "work in the browser. Please install it using: pip install pyodide-http"
            )

        pyodide_http.patch_all()


class HTTPAdapter(FSSpecAdapter):
    def supports(self, url: str) -> bool:
        return url.startswith("http://") or url.startswith("https://")

    def _get_filesystem(
        self, url: str, no_cache: bool = False
    ) -> fsspec.AbstractFileSystem:
        try:
            import aiohttp
        except ImportError:
            raise AdapterError(
                "Seems like you don't have `aiohttp` installed, which is required to authenticate. "
                "Please install it using: pip install aiohttp"
            )

        check_requests_patch()

        basic_authentication = get_config("adapters.http.basic_authentication")

        client_kwargs = {}
        if basic_authentication:
            client_kwargs["auth"] = aiohttp.BasicAuth(*basic_authentication)

        if no_cache:
            return fsspec.filesystem("http", client_kwargs=client_kwargs)
        else:
            return fsspec.filesystem(
                "simplecache",
                target_protocol="http",
                target_options={"client_kwargs": client_kwargs},
                cache_storage=get_config("cache"),
            )

    def is_directory(self, url: str) -> bool:
        """
        Check if the given URL points to a directory.
        """
        fs = self._get_filesystem(url, no_cache=True)
        return fs.isdir(url)



"""XML parser for Stats Perform MA1 feeds."""

from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, Tuple

from kloppy.domain import Ground, Period, Player, PositionType, Score, Team
from kloppy.exceptions import DeserializationError

from ..formation_mapping import (
    FormationType,
    formation_name_mapping,
    formation_position_mapping,
)
from .base import OptaXMLParser


class MA1XMLParser(OptaXMLParser):
    """Extract data from a Stats Perform MA1 data stream."""

    def extract_periods(self) -> List[Period]:
        parsed_periods = []
        live_data = self.root.liveData
        match_details = live_data.matchDetails
        periods = match_details.periods
        for period in periods.iterchildren(tag="period"):
            parsed_periods.append(
                Period(
                    id=int(period.get("id")),
                    start_timestamp=datetime.strptime(
                        period.get("start"), "%Y-%m-%dT%H:%M:%SZ"
                    ).replace(tzinfo=timezone.utc),
                    end_timestamp=datetime.strptime(
                        period.get("end"), "%Y-%m-%dT%H:%M:%SZ"
                    ).replace(tzinfo=timezone.utc),
                )
            )
        return parsed_periods

    def extract_score(self) -> Optional[Score]:
        return None

    def extract_lineups(self) -> Tuple[Team, Team]:
        teams = {}
        for parsed_team in self._parse_teams():
            team_id = parsed_team["team_id"]
            teams[team_id] = Team(
                team_id=team_id,
                name=parsed_team["name"],
                ground=(
                    Ground.HOME
                    if parsed_team["ground"] == "home"
                    else Ground.AWAY
                ),
            )

        for parsed_player in self._parse_players():
            player_id = parsed_player["player_id"]
            team_id = parsed_player["team_id"]
            team = teams[team_id]
            player = Player(
                player_id=player_id,
                team=team,
                jersey_no=parsed_player["jersey_no"],
                name=parsed_player["name"],
                first_name=parsed_player["first_name"],
                last_name=parsed_player["last_name"],
                starting=parsed_player["starting"],
                starting_position=parsed_player["position"],
            )
            team.players.append(player)

        home_team = next(
            (team for team in teams.values() if team.ground == Ground.HOME),
            None,
        )
        away_team = next(
            (team for team in teams.values() if team.ground == Ground.AWAY),
            None,
        )
        if home_team is None:
            raise DeserializationError("Could not find home team")
        if away_team is None:
            raise DeserializationError("Could not find away team")
        if len(home_team.players) == 0 or len(away_team.players) == 0:
            raise DeserializationError("Lineup incomplete")
        return home_team, away_team

    def _parse_teams(self) -> List[Dict[str, Any]]:
        parsed_teams = []
        match_info = self.root.matchInfo
        teams = match_info.contestants.iterchildren(tag="contestant")

        team_formations = {}
        live_data = self.root.liveData
        line_ups = live_data.iterchildren(tag="lineUp")
        for line_up in line_ups:
            team_id = line_up.get("contestantId")
            raw_formation = line_up.get("formationUsed")
            formation = formation_name_mapping.get(
                raw_formation, FormationType.UNKNOWN
            )
            team_formations[team_id] = formation

        for team in teams:
            team_attributes = team.attrib
            team_id = team_attributes["id"]
            parsed_teams.append(
                {
                    "team_id": team_id,
                    "name": team_attributes["name"],
                    "ground": team_attributes["position"],
                    "starting_formation": team_formations[team_id],
                }
            )
        return parsed_teams

    def _parse_players(self) -> List[Dict[str, Any]]:
        parsed_players = []
        live_data = self.root.liveData
        line_ups = live_data.iterchildren(tag="lineUp")
        for line_up in line_ups:
            team_id = line_up.get("contestantId")
            raw_formation = line_up.get("formationUsed")
            formation = formation_name_mapping.get(
                raw_formation, FormationType.UNKNOWN
            )

            players = line_up.iterchildren(tag="player")
            for player in players:
                player_attributes = player.attrib
                player_id = player_attributes["playerId"]

                if "formationPlace" in player_attributes:
                    player_position = (
                        formation_position_mapping[formation][
                            int(player_attributes["formationPlace"])
                        ]
                        if formation != FormationType.UNKNOWN
                        else PositionType.Unknown
                    )
                    starting = True
                else:
                    player_position = None
                    starting = False
                parsed_players.append(
                    {
                        "player_id": player_id,
                        "team_id": team_id,
                        "jersey_no": int(player_attributes["shirtNumber"]),
                        "name": player_attributes["matchName"],
                        "first_name": player_attributes["shortFirstName"]
                        if "shortFirstName" in player_attributes
                        else player_attributes["firstName"],
                        "last_name": player_attributes["shortLastName"]
                        if "shortLastName" in player_attributes
                        else player_attributes["lastName"],
                        "starting": starting,
                        "position": player_position,
                    }
                )
        return parsed_players



from __future__ import annotations

import sys
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field, replace
from datetime import datetime, timedelta
from enum import Enum, Flag
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Literal,
    Optional,
    TypeVar,
    Union,
    overload,
)

from kloppy.utils import deprecated, snake_case

if TYPE_CHECKING:
    from ..services.transformers.data_record import (
        Column,
        NamedColumns,
    )

from .position import PositionType

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

if sys.version_info >= (3, 11):
    from typing import Self, Unpack
else:
    from typing_extensions import Self, Unpack

from kloppy.exceptions import (
    InvalidFilterError,
    KloppyParameterError,
    OrientationError,
)

from .formation import FormationType
from .pitch import (
    DEFAULT_PITCH_LENGTH,
    DEFAULT_PITCH_WIDTH,
    Dimension,
    ImperialPitchDimensions,
    MetricPitchDimensions,
    NormalizedPitchDimensions,
    OptaPitchDimensions,
    PitchDimensions,
    Unit,
    WyscoutPitchDimensions,
)
from .time import Period, Time, TimeContainer


@dataclass
class Score:
    """
    The scoreline of a match.

    Attributes:
        home: Goals scored by the home team.
        away: Goals scored by the away team.
    """

    home: int
    away: int


class Ground(Enum):
    """
    Whether a team is playing at home or away.

    Attributes:
        HOME (Ground): The team is playing at home.
        AWAY (Ground): The team is playing away.
        REFEREE (Ground): Referee (could be used in tracking data).
    """

    HOME = "home"
    AWAY = "away"
    REFEREE = "referee"

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


class Provider(Enum):
    """
    Data providers.

    Attributes:
        METRICA (Provider):
        TRACAB (Provider):
        SECONDSPECTRUM (Provider):
        OPTA (Provider):
        PFF (Provider):
        SKILLCORNER (Provider):
        STATSBOMB (Provider):
        SPORTEC (Provider):
        WYSCOUT (Provider):
        KLOPPY (Provider):
        DATAFACTORY (Provider):
        STATSPERFORM (Provider):
        SPORTVU (Provider):
        OTHER (Provider):
    """

    METRICA = "metrica"
    TRACAB = "tracab"
    SECONDSPECTRUM = "second_spectrum"
    OPTA = "opta"
    PFF = "pff"
    SKILLCORNER = "skillcorner"
    STATSBOMB = "statsbomb"
    SPORTEC = "sportec"
    WYSCOUT = "wyscout"
    KLOPPY = "kloppy"
    DATAFACTORY = "datafactory"
    STATSPERFORM = "statsperform"
    HAWKEYE = "hawkeye"
    SPORTVU = "sportvu"
    SIGNALITY = "signality"
    OTHER = "other"

    def __str__(self):
        return self.value


class OfficialType(Enum):
    """Enumeration for types of officials (referees)."""

    VideoAssistantReferee = "Video Assistant Referee"
    AssistantVideoAssistantReferee = "Assistant Video Assistant Referee"
    MainReferee = "Main Referee"
    AssistantReferee = "Assistant Referee"
    FourthOfficial = "Fourth Official"
    Unknown = "Unknown"

    def __str__(self):
        return self.value


@dataclass(frozen=True)
class Official:
    """
    Represents an official (referee) with optional names and roles.
    """

    official_id: str
    name: Optional[str] = None
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    role: Optional[OfficialType] = None

    @property
    def full_name(self):
        """
        Returns the full name of the official, falling back to role-based or ID-based naming.
        """
        if self.name:
            return self.name
        if self.first_name and self.last_name:
            return f"{self.first_name} {self.last_name}"
        if self.last_name:
            return self.last_name
        if self.role:
            return f"{snake_case(str(self.role))}_{self.official_id}"
        return f"official_{self.official_id}"


@dataclass(frozen=True)
class Player:
    """
    A player in a team.

    Attributes:
        player_id (str): Identifier given by the provider.
        team (team): The player's team.
        jersey_no (int): The player's jersey number.
        first_name (str, optional): The player's first name.
        last_name (str, optional): The player's last name.
        name (str, optional): Full name of the player.
        full_name (str): If `name` is not set, this will be the concatenation
            of `first_name` and `last_name` or if these are also not set,
            the concatenation of the team's ground and the jersey number.
        starting (bool): `True` when player is part of the starting XI.
        starting_position (Position, optional): The player's starting position
            or `None` if the player is not starting.
        poisitions (TimeContainer[Position]): The player's positions over time.
        attributes (dict): Additional attributes given by the provider.
    """

    player_id: str
    team: "Team"
    jersey_no: int
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    name: Optional[str] = None

    # match specific
    starting: bool = False
    starting_position: Optional[PositionType] = None
    positions: TimeContainer[PositionType] = field(
        default_factory=TimeContainer, compare=False
    )

    attributes: Dict = field(default_factory=dict, compare=False)

    @property
    def full_name(self):
        if self.name:
            return self.name
        if self.first_name or self.last_name:
            return f"{self.first_name} {self.last_name}"
        return f"{self.team.ground}_{self.jersey_no}"

    @property
    @deprecated("starting_position or positions should be used")
    def position(self) -> Optional[PositionType]:
        try:
            return self.positions.last()
        except KeyError:
            return None

    def __str__(self):
        return self.full_name

    def __hash__(self):
        return hash(self.player_id)

    def __eq__(self, other):
        if not isinstance(other, Player):
            return False
        return self.player_id == other.player_id

    def set_position(self, time: Time, position: Optional[PositionType]):
        self.positions.set(time, position)


@dataclass
class Team:
    """
    A team in a match.

    Attributes:
        team_id (str): Identifier given by the provider.
        name (str): Readable name of the team.
        ground (Ground): The team's ground (home or away).
        players (List[Player]): The team's players.
        starting_formation (FormationType, optional): The team's starting formation.
    """

    team_id: str
    name: str
    ground: Ground
    starting_formation: Optional[FormationType] = None
    formations: TimeContainer[FormationType] = field(
        default_factory=TimeContainer, compare=False
    )
    players: List[Player] = field(default_factory=list)

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(self.team_id)

    def __eq__(self, other):
        if not isinstance(other, Team):
            return False
        return self.team_id == other.team_id

    def get_player_by_jersey_number(self, jersey_no: int) -> Optional[Player]:
        """Get a player by their jersey number.

        Args:
            jersey_no (int): The jersey number of the player.

        Returns:
            Player: The player with the given jersey number or `None` if no
                    player with that jersey number is found.
        """
        jersey_no = int(jersey_no)
        for player in self.players:
            if player.jersey_no == jersey_no:
                return player

        return None

    def get_player_by_position(
        self, position: PositionType, time: Time
    ) -> Optional[Player]:
        """Get a player by their position at a given time.

        Args:
            position (PositionType): The position.
            time (Time): The time for which the position should be retrieved.

        Returns:
            Player: The player with the given position or `None` if no player
                    with that position is found.
        """
        for player in self.players:
            if player.positions.items:
                try:
                    player_position = player.positions.value_at(time)
                except KeyError:  # player that is subbed in later
                    continue
                if player_position and player_position == position:
                    return player

        return None

    def get_player_by_id(self, player_id: Union[int, str]) -> Optional[Player]:
        """Get a player by their identifier.

        Args:
            player_id (int or str): The identifier of the player.

        Returns:
            Player: The player with the given identifier or `None` if no player
                    with that identifier is found.
        """
        player_id = str(player_id)

        for player in self.players:
            if player.player_id == player_id:
                return player

        return None

    def set_formation(self, time: Time, formation: Optional[FormationType]):
        self.formations.set(time, formation)


class BallState(Enum):
    """
    Whether the ball is in play or not.

    Attributes:
        ALIVE (BallState): Ball is in play.
        DEAD (BallState): Ball is not in play.
    """

    ALIVE = "alive"
    DEAD = "dead"

    def __repr__(self):
        return self.value


class Orientation(Enum):
    """
    The attacking direction of each team in a dataset.

    Attributes:
        BALL_OWNING_TEAM: The team that is currently in possession of the ball
            plays from left to right.
        ACTION_EXECUTING_TEAM: The team that executes the action
            plays from left to right. Used in event stream data only. Equivalent
            to "BALL_OWNING_TEAM" for tracking data.
        HOME_AWAY: The home team plays from left to right in the first period.
            The away team plays from left to right in the second period.
        AWAY_HOME: The away team plays from left to right in the first period.
            The home team plays from left to right in the second period.
        STATIC_HOME_AWAY: The home team plays from left to right in both periods.
        STATIC_AWAY_HOME: The away team plays from left to right in both periods.
        NOT_SET: The attacking direction is not defined.

    Notes:
        The attacking direction is not defined for penalty shootouts in the
        `HOME_AWAY`, `AWAY_HOME`, `STATIC_HOME_AWAY`, and `STATIC_AWAY_HOME`
        orientations. This period is ignored in orientation transforms
        involving one of these orientations and keeps its original
        attacking direction.
    """

    # change when possession changes
    BALL_OWNING_TEAM = "ball-owning-team"

    # depends on team which executed the action
    ACTION_EXECUTING_TEAM = "action-executing-team"

    # changes during half-time
    HOME_AWAY = "home-away"
    AWAY_HOME = "away-home"

    # won't change during match
    STATIC_HOME_AWAY = "fixed-home-away"
    STATIC_AWAY_HOME = "fixed-away-home"

    # Not set in dataset
    NOT_SET = "not-set"

    def __repr__(self):
        return self.value


class AttackingDirection(Enum):
    """
    The direction of play for the home team.

    Attributes:
        LTR (AttackingDirection): Home team is playing from left to right
        RTL (AttackingDirection): Home team is playing from right to left
        NOT_SET (AttackingDirection): not set yet
    """

    LTR = "left-to-right"
    RTL = "right-to-left"
    NOT_SET = "not-set"

    @staticmethod
    def from_orientation(
        orientation: Orientation,
        period: Optional[Period] = None,
        ball_owning_team: Optional[Team] = None,
        action_executing_team: Optional[Team] = None,
    ) -> "AttackingDirection":
        """Determines the attacking direction for a specific data record.

        Args:
            orientation: The orientation of the dataset.
            period: The period of the data record.
            ball_owning_team: The team that is in possession of the ball.
            action_executing_team: The team that executes the action.

        Raises:
            OrientationError: If the attacking direction cannot be determined
                from the given data.

        Returns:
            The attacking direction for the given data record.
        """
        if orientation == Orientation.STATIC_HOME_AWAY:
            return AttackingDirection.LTR
        if orientation == Orientation.STATIC_AWAY_HOME:
            return AttackingDirection.RTL
        if orientation == Orientation.HOME_AWAY:
            if period is None:
                raise OrientationError(
                    "You must provide a period to determine the attacking direction"
                )
            dirmap = {
                1: AttackingDirection.LTR,
                2: AttackingDirection.RTL,
                3: AttackingDirection.LTR,
                4: AttackingDirection.RTL,
            }
            if period.id in dirmap:
                return dirmap[period.id]
            raise OrientationError(
                "This orientation is not defined for period %s" % period.id
            )
        if orientation == Orientation.AWAY_HOME:
            if period is None:
                raise OrientationError(
                    "You must provide a period to determine the attacking direction"
                )
            dirmap = {
                1: AttackingDirection.RTL,
                2: AttackingDirection.LTR,
                3: AttackingDirection.RTL,
                4: AttackingDirection.LTR,
            }
            if period.id in dirmap:
                return dirmap[period.id]
            raise OrientationError(
                "This orientation is not defined for period %s" % period.id
            )
        if orientation == Orientation.BALL_OWNING_TEAM:
            if ball_owning_team is None:
                raise OrientationError(
                    "You must provide the ball owning team to determine the attacking direction"
                )
            if ball_owning_team is not None:
                if ball_owning_team.ground == Ground.HOME:
                    return AttackingDirection.LTR
                if ball_owning_team.ground == Ground.AWAY:
                    return AttackingDirection.RTL
                raise OrientationError(
                    "Invalid ball_owning_team: %s", ball_owning_team
                )
            return AttackingDirection.NOT_SET
        if orientation == Orientation.ACTION_EXECUTING_TEAM:
            if action_executing_team is None:
                raise ValueError(
                    "You must provide the action executing team to determine the attacking direction"
                )
            if action_executing_team.ground == Ground.HOME:
                return AttackingDirection.LTR
            if action_executing_team.ground == Ground.AWAY:
                return AttackingDirection.RTL
            raise OrientationError(
                "Invalid action_executing_team: %s", action_executing_team
            )
        raise OrientationError("Unknown orientation: %s", orientation)

    def __repr__(self):
        return self.value


class VerticalOrientation(Enum):
    """
    The orientation of the y-axis in a [`CoordinateSystem`][kloppy.domain.CoordinateSystem].

    Attributes:
        TOP_TO_BOTTOM: The y coordinate increases from top to bottom of the pitch.
        BOTTOM_TO_TOP: The y coordinate decreases from top to bottom of the pitch.
    """

    TOP_TO_BOTTOM = "top-to-bottom"
    BOTTOM_TO_TOP = "bottom-to-top"


class Origin(Enum):
    """
    The location of the origin in a [`CoordinateSystem`][kloppy.domain.CoordinateSystem].

    Defines where the (0, 0) point is located on the field.

    Attributes:
        TOP_LEFT: Origin at the top left of the field
        BOTTOM_LEFT: Origin at the bottom left of the field
        CENTER: Origin at the center of the field
    """

    TOP_LEFT = "top-left"
    BOTTOM_LEFT = "bottom-left"
    CENTER = "center"

    def __str__(self):
        return self.value


class CoordinateSystem(ABC):
    """
    Base class for coordinate systems.

    A coordinate system defines how coordinates are represented in a dataset.

    Attributes:
        origin (Origin): The location of the origin.
        vertical_orientation (VerticalOrientation): The orientation of the y-axis.
        pitch_dimensions (PitchDimensions): The dimensions of the pitch.
        normalized (bool): Whether the pitch dimensions are normalized. This
            means that the coordinates are mapped to a fixed range, e.g. from
            0 to 1. In contrast, non-normalized coordinates correspond to the
            real-world dimensions of the pitch.
        pitch_length (float, optional): The real length of the pitch.
        pitch_width (float, optional): The real width of the pitch.
    """

    @property
    @abstractmethod
    def origin(self) -> Origin:
        raise NotImplementedError

    @property
    @abstractmethod
    def vertical_orientation(self) -> VerticalOrientation:
        raise NotImplementedError

    @property
    @abstractmethod
    def pitch_dimensions(self) -> PitchDimensions:
        raise NotImplementedError

    @property
    def normalized(self) -> bool:
        return isinstance(self.pitch_dimensions, NormalizedPitchDimensions)

    @property
    def pitch_length(self) -> Optional[float]:
        return self.pitch_dimensions.pitch_length

    @property
    def pitch_width(self) -> Optional[float]:
        return self.pitch_dimensions.pitch_width

    def __eq__(self, other):
        if isinstance(other, CoordinateSystem):
            return (
                self.origin == other.origin
                and self.vertical_orientation == other.vertical_orientation
                and self.pitch_dimensions == other.pitch_dimensions
            )

        return False

    def to_mplsoccer(self):
        """ "Convert the coordinate system to a mplsoccer BaseDims object.

        Example:
        >>> from kloppy.domain import KloppyCoordinateSystem
        >>> from mplsoccer import Pitch
        >>> coordinate_system = KloppyCoordinateSystem()
        >>> pitch = Pitch(dimensions=coordinate_system.to_mplsoccer())

        Note:
            This method is experimental and may change in the future.
        """
        try:
            from mplsoccer.dimensions import BaseDims
        except ImportError:
            raise ImportError(
                "Seems like you don't have `mplsoccer` installed. "
                "Please install it using: pip install mplsoccer"
            )

        if (
            self.pitch_dimensions.x_dim.min is None
            or self.pitch_dimensions.x_dim.max is None
        ):
            raise ValueError(
                "The x-dimensions of the pitch must be fully defined."
            )
        if (
            self.pitch_dimensions.y_dim.min is None
            or self.pitch_dimensions.y_dim.max is None
        ):
            raise ValueError(
                "The y-dimensions of the pitch must be fully defined."
            )

        pitch_length = (
            self.pitch_dimensions.pitch_length or DEFAULT_PITCH_LENGTH
        )
        pitch_width = self.pitch_dimensions.pitch_width or DEFAULT_PITCH_WIDTH

        invert_y = (
            self.vertical_orientation == VerticalOrientation.TOP_TO_BOTTOM
        )
        origin_center = self.origin == Origin.CENTER

        neg_if_inverted = -1 if invert_y else 1
        center = (0, 0)
        if self.origin == Origin.BOTTOM_LEFT:
            center = (
                (
                    self.pitch_dimensions.x_dim.max
                    - self.pitch_dimensions.x_dim.min
                )
                / 2,
                (
                    self.pitch_dimensions.y_dim.max
                    - self.pitch_dimensions.y_dim.min
                )
                / 2,
            )
        elif self.origin == Origin.TOP_LEFT:
            neg_if_inverted = -1
            center = (
                (
                    self.pitch_dimensions.x_dim.max
                    - self.pitch_dimensions.x_dim.min
                )
                / 2,
                (
                    self.pitch_dimensions.y_dim.max
                    - self.pitch_dimensions.y_dim.min
                )
                / 2,
            )

        dim = BaseDims(
            left=self.pitch_dimensions.x_dim.min,
            right=self.pitch_dimensions.x_dim.max,
            bottom=self.pitch_dimensions.y_dim.min
            if not invert_y
            else self.pitch_dimensions.y_dim.max,
            top=self.pitch_dimensions.y_dim.max
            if not invert_y
            else self.pitch_dimensions.y_dim.min,
            width=self.pitch_dimensions.x_dim.max
            - self.pitch_dimensions.x_dim.min,
            length=self.pitch_dimensions.y_dim.max
            - self.pitch_dimensions.y_dim.min,
            goal_bottom=center[1]
            - (neg_if_inverted / 2 * self.pitch_dimensions.goal_width),
            goal_top=center[1]
            + (neg_if_inverted / 2 * self.pitch_dimensions.goal_width),
            six_yard_left=self.pitch_dimensions.x_dim.min
            + self.pitch_dimensions.six_yard_length,
            six_yard_right=self.pitch_dimensions.x_dim.max
            - self.pitch_dimensions.six_yard_length,
            six_yard_bottom=center[1]
            - (neg_if_inverted / 2 * self.pitch_dimensions.six_yard_width),
            six_yard_top=center[1]
            + (neg_if_inverted / 2 * self.pitch_dimensions.six_yard_width),
            penalty_spot_distance=self.pitch_dimensions.penalty_spot_distance,
            penalty_left=self.pitch_dimensions.x_dim.min
            + self.pitch_dimensions.penalty_spot_distance,
            penalty_right=self.pitch_dimensions.x_dim.max
            - self.pitch_dimensions.penalty_spot_distance,
            penalty_area_left=self.pitch_dimensions.x_dim.min
            + self.pitch_dimensions.penalty_area_length,
            penalty_area_right=self.pitch_dimensions.x_dim.max
            - self.pitch_dimensions.penalty_area_length,
            penalty_area_bottom=center[1]
            - (neg_if_inverted / 2 * self.pitch_dimensions.penalty_area_width),
            penalty_area_top=center[1]
            + (neg_if_inverted / 2 * self.pitch_dimensions.penalty_area_width),
            center_width=center[1],
            center_length=center[0],
            goal_width=self.pitch_dimensions.goal_width,
            goal_length=self.pitch_dimensions.goal_height,
            six_yard_width=self.pitch_dimensions.six_yard_width,
            six_yard_length=self.pitch_dimensions.six_yard_length,
            penalty_area_width=self.pitch_dimensions.penalty_area_width,
            penalty_area_length=self.pitch_dimensions.penalty_area_length,
            circle_diameter=self.pitch_dimensions.circle_radius * 2,
            corner_diameter=self.pitch_dimensions.corner_radius * 2,
            arc=0,
            invert_y=invert_y,
            origin_center=origin_center,
            pad_default=0.02
            * (
                self.pitch_dimensions.x_dim.max
                - self.pitch_dimensions.x_dim.min
            ),
            pad_multiplier=1,
            aspect_equal=False
            if self.pitch_dimensions.unit == Unit.NORMED
            else True,
            pitch_width=pitch_width,
            pitch_length=pitch_length,
            aspect=pitch_width / pitch_length
            if self.pitch_dimensions.unit == Unit.NORMED
            else 1.0,
        )
        return dim


class ProviderCoordinateSystem(CoordinateSystem):
    def __init__(
        self,
        pitch_length: Optional[float] = None,
        pitch_width: Optional[float] = None,
    ):
        self._pitch_length = pitch_length
        self._pitch_width = pitch_width

    @property
    @abstractmethod
    def provider(self) -> Provider:
        raise NotImplementedError


class CustomCoordinateSystem(CoordinateSystem):
    def __init__(
        self,
        origin: Origin,
        vertical_orientation: VerticalOrientation,
        pitch_dimensions: PitchDimensions,
    ):
        self._origin = origin
        self._vertical_orientation = vertical_orientation
        self._pitch_dimensions = pitch_dimensions

    @property
    def origin(self) -> Origin:
        return self._origin

    @property
    def vertical_orientation(self) -> VerticalOrientation:
        return self._vertical_orientation

    @property
    def pitch_dimensions(self) -> PitchDimensions:
        return self._pitch_dimensions


class KloppyCoordinateSystem(ProviderCoordinateSystem):
    """
    Kloppy's default coordinate system.

    Uses a normalized pitch with the origin at the top left and the y-axis
    oriented from top to bottom. The coordinates range from 0 to 1.

    If no pitch length and width are provided, the default pitch dimensions
    are 105m x 68m.
    """

    @property
    def provider(self) -> Provider:
        return Provider.KLOPPY

    @property
    def origin(self) -> Origin:
        return Origin.TOP_LEFT

    @property
    def vertical_orientation(self) -> VerticalOrientation:
        return VerticalOrientation.TOP_TO_BOTTOM

    @property
    def pitch_dimensions(self) -> PitchDimensions:
        if self._pitch_length is not None and self._pitch_width is not None:
            return NormalizedPitchDimensions(
                x_dim=Dimension(0, 1),
                y_dim=Dimension(0, 1),
                pitch_length=self._pitch_length,
                pitch_width=self._pitch_width,
                standardized=False,
            )
        else:
            return NormalizedPitchDimensions(
                x_dim=Dimension(0, 1),
                y_dim=Dimension(0, 1),
                pitch_length=105,
                pitch_width=68,
                standardized=True,
            )


class MetricaCoordinateSystem(KloppyCoordinateSystem):
    """
    Metrica coordinate system.

    Uses a normalized pitch with the origin at the top left and the y-axis
    oriented from top to bottom. The coordinates range from 0 to 1.

    If no pitch length and width are provided, the default pitch dimensions
    are 105m x 68m.

    Notes:
        The Metrica coordinate system is the same as the
        [`KloppyCoordinateSystem`][kloppy.domain.KloppyCoordinateSystem].
    """

    @property
    def provider(self) -> Provider:
        return Provider.METRICA


class TracabCoordinateSystem(ProviderCoordinateSystem):
    """
    Tracab coordinate system.

    Uses a pitch with the origin at the center and the y-axis oriented from
    bottom to top. The coordinates are in centimeters.
    """

    @property
    def provider(self) -> Provider:
        return Provider.TRACAB

    @property
    def origin(self) -> Origin:
        return Origin.CENTER

    @property
    def vertical_orientation(self) -> VerticalOrientation:
        return VerticalOrientation.BOTTOM_TO_TOP

    @property
    def pitch_dimensions(self) -> PitchDimensions:
        if self._pitch_length is not None and self._pitch_width is not None:
            return MetricPitchDimensions(
                x_dim=Dimension(
                    -1 * self._pitch_length / 2, self._pitch_length / 2
                ),
                y_dim=Dimension(
                    -1 * self._pitch_width / 2, self._pitch_width / 2
                ),
                pitch_length=self._pitch_length,
                pitch_width=self._pitch_width,
                standardized=False,
            ).convert(to_unit=Unit.CENTIMETERS)
        else:
            return MetricPitchDimensions(
                x_dim=Dimension(None, None),
                y_dim=Dimension(None, None),
                pitch_length=None,
                pitch_width=None,
                standardized=False,
            ).convert(to_unit=Unit.CENTIMETERS)


class SecondSpectrumCoordinateSystem(ProviderCoordinateSystem):
    """
    Second Spectrum coordinate system.

    Uses a pitch with the origin at the center and the y-axis oriented from
    bottom to top. The coordinates are in meters.
    """

    @property
    def provider(self) -> Provider:
        return Provider.SECONDSPECTRUM

    @property
    def origin(self) -> Origin:
        return Origin.CENTER

    @property
    def vertical_orientation(self) -> VerticalOrientation:
        return VerticalOrientation.BOTTOM_TO_TOP

    @property
    def pitch_dimensions(self) -> PitchDimensions:
        if self._pitch_length is not None and self._pitch_width is not None:
            return MetricPitchDimensions(
                x_dim=Dimension(
                    -1 * self._pitch_length / 2, self._pitch_length / 2
                ),
                y_dim=Dimension(
                    -1 * self._pitch_width / 2, self._pitch_width / 2
                ),
                pitch_length=self._pitch_length,
                pitch_width=self._pitch_width,
                standardized=False,
            )
        else:
            return MetricPitchDimensions(
                x_dim=Dimension(None, None),
                y_dim=Dimension(None, None),
                pitch_length=None,
                pitch_width=None,
                standardized=False,
            )


class OptaCoordinateSystem(ProviderCoordinateSystem):
    """
    Opta coordinate system.

    Uses a normalized pitch with the origin at the bottom left and the y-axis
    oriented from bottom to top. The coordinates range from 0 to 100.
    """

    @property
    def provider(self) -> Provider:
        return Provider.OPTA

    @property
    def origin(self) -> Origin:
        return Origin.BOTTOM_LEFT

    @property
    def vertical_orientation(self) -> VerticalOrientation:
        return VerticalOrientation.BOTTOM_TO_TOP

    @property
    def pitch_dimensions(self) -> PitchDimensions:
        return OptaPitchDimensions(
            pitch_length=self._pitch_length, pitch_width=self._pitch_width
        )


class SportecEventDataCoordinateSystem(ProviderCoordinateSystem):
    """
    Sportec event data coordinate system.

    Uses a pitch with the origin at the bottom left and the y-axis oriented
    from top to bottom. The coordinates are in meters.
    """

    @property
    def provider(self) -> Provider:
        return Provider.SPORTEC

    @property
    def origin(self) -> Origin:
        return Origin.BOTTOM_LEFT

    @property
    def vertical_orientation(self) -> VerticalOrientation:
        return VerticalOrientation.BOTTOM_TO_TOP

    @property
    def pitch_dimensions(self) -> PitchDimensions:
        return MetricPitchDimensions(
            x_dim=Dimension(0, self._pitch_length),
            y_dim=Dimension(0, self._pitch_width),
            pitch_length=self._pitch_length,
            pitch_width=self._pitch_width,
            standardized=False,
        )


class SportecTrackingDataCoordinateSystem(ProviderCoordinateSystem):
    """
    Sportec tracking data coordinate system.

    Uses a pitch with the origin at the center and the y-axis oriented
    from bottom to top. The coordinates are in meters.
    """

    @property
    def provider(self) -> Provider:
        return Provider.SPORTEC

    @property
    def origin(self) -> Origin:
        return Origin.CENTER

    @property
    def vertical_orientation(self) -> VerticalOrientation:
        return VerticalOrientation.BOTTOM_TO_TOP

    @property
    def pitch_dimensions(self) -> PitchDimensions:
        if self._pitch_length is not None and self._pitch_width is not None:
            return MetricPitchDimensions(
                x_dim=Dimension(
                    -1 * self._pitch_length / 2, self._pitch_length / 2
                ),
                y_dim=Dimension(
                    -1 * self._pitch_width / 2, self._pitch_width / 2
                ),
                pitch_length=self._pitch_length,
                pitch_width=self._pitch_width,
                standardized=False,
            )
        else:
            return MetricPitchDimensions(
                x_dim=Dimension(None, None),
                y_dim=Dimension(None, None),
                pitch_length=None,
                pitch_width=None,
                standardized=False,
            )


class StatsBombCoordinateSystem(ProviderCoordinateSystem):
    """
    StatsBomb coordinate system.

    Uses a normalized pitch with the origin at the top left and the y-axis
    oriented from top to bottom. The x-coordinates range from 0 to 120 and
    the y-coordinates range from 0 to 80.
    """

    @property
    def provider(self) -> Provider:
        return Provider.STATSBOMB

    @property
    def origin(self) -> Origin:
        return Origin.TOP_LEFT

    @property
    def vertical_orientation(self) -> VerticalOrientation:
        return VerticalOrientation.TOP_TO_BOTTOM

    @property
    def pitch_dimensions(self) -> PitchDimensions:
        return ImperialPitchDimensions(
            x_dim=Dimension(0, 120),
            y_dim=Dimension(0, 80),
            pitch_length=self._pitch_length,
            pitch_width=self._pitch_width,
            standardized=True,
        )


class PFFCoordinateSystem(ProviderCoordinateSystem):
    """
    PFF coordinate system.

    Uses a pitch with the origin at the center and the y-axis oriented
    from bottom to top. The coordinates are in meters.
    """

    @property
    def provider(self) -> Provider:
        return Provider.PFF

    @property
    def origin(self) -> Origin:
        return Origin.CENTER

    @property
    def vertical_orientation(self) -> VerticalOrientation:
        return VerticalOrientation.BOTTOM_TO_TOP

    @property
    def pitch_dimensions(self) -> PitchDimensions:
        if self._pitch_length is not None and self._pitch_width is not None:
            return MetricPitchDimensions(
                x_dim=Dimension(
                    -1 * self._pitch_length / 2, self._pitch_length / 2
                ),
                y_dim=Dimension(
                    -1 * self._pitch_width / 2, self._pitch_width / 2
                ),
                pitch_length=self._pitch_length,
                pitch_width=self._pitch_width,
                standardized=False,
            )
        else:
            return MetricPitchDimensions(
                x_dim=Dimension(None, None),
                y_dim=Dimension(None, None),
                pitch_length=None,
                pitch_width=None,
                standardized=False,
            )


class WyscoutCoordinateSystem(ProviderCoordinateSystem):
    """
    Wyscout coordinate system.

    Uses a normalized pitch with the origin at the top left and the y-axis
    oriented from top to bottom. The coordinates range from 0 to 100.
    """

    @property
    def provider(self) -> Provider:
        return Provider.WYSCOUT

    @property
    def origin(self) -> Origin:
        return Origin.TOP_LEFT

    @property
    def vertical_orientation(self) -> VerticalOrientation:
        return VerticalOrientation.TOP_TO_BOTTOM

    @property
    def pitch_dimensions(self) -> PitchDimensions:
        return WyscoutPitchDimensions(
            pitch_length=self._pitch_length, pitch_width=self._pitch_width
        )


class SkillCornerCoordinateSystem(ProviderCoordinateSystem):
    """
    SkillCorner coordinate system.

    Uses a pitch with the origin at the center and the y-axis oriented
    from bottom to top. The coordinates are in meters.
    """

    @property
    def provider(self) -> Provider:
        return Provider.SKILLCORNER

    @property
    def origin(self) -> Origin:
        return Origin.CENTER

    @property
    def vertical_orientation(self) -> VerticalOrientation:
        return VerticalOrientation.BOTTOM_TO_TOP

    @property
    def pitch_dimensions(self) -> PitchDimensions:
        if self._pitch_length is not None and self._pitch_width is not None:
            return MetricPitchDimensions(
                x_dim=Dimension(
                    -1 * self._pitch_length / 2, self._pitch_length / 2
                ),
                y_dim=Dimension(
                    -1 * self._pitch_width / 2, self._pitch_width / 2
                ),
                pitch_length=self._pitch_length,
                pitch_width=self._pitch_width,
                standardized=False,
            )
        else:
            return MetricPitchDimensions(
                x_dim=Dimension(None, None),
                y_dim=Dimension(None, None),
                pitch_length=None,
                pitch_width=None,
                standardized=False,
            )


class SignalityCoordinateSystem(ProviderCoordinateSystem):
    @property
    def provider(self) -> Provider:
        return Provider.SIGNALITY

    @property
    def origin(self) -> Origin:
        return Origin.CENTER

    @property
    def vertical_orientation(self) -> VerticalOrientation:
        return VerticalOrientation.BOTTOM_TO_TOP

    @property
    def pitch_dimensions(self) -> PitchDimensions:
        if self._pitch_length is not None and self._pitch_width is not None:
            return MetricPitchDimensions(
                x_dim=Dimension(
                    -1 * self._pitch_length / 2, self._pitch_length / 2
                ),
                y_dim=Dimension(
                    -1 * self._pitch_width / 2, self._pitch_width / 2
                ),
                pitch_length=self._pitch_length,
                pitch_width=self._pitch_width,
                standardized=False,
            )
        else:
            return MetricPitchDimensions(
                x_dim=Dimension(None, None),
                y_dim=Dimension(None, None),
                pitch_length=None,
                pitch_width=None,
                standardized=False,
            )


class DatafactoryCoordinateSystem(ProviderCoordinateSystem):
    """
    Datafactory coordinate system.

    Uses a normalized pitch with the origin at the top left and the y-axis
    oriented from top to bottom. The coordinates range from -1 to 1.
    """

    @property
    def provider(self) -> Provider:
        return Provider.DATAFACTORY

    @property
    def origin(self) -> Origin:
        return Origin.CENTER

    @property
    def vertical_orientation(self) -> VerticalOrientation:
        return VerticalOrientation.TOP_TO_BOTTOM

    @property
    def pitch_dimensions(self) -> PitchDimensions:
        if self._pitch_length is not None and self._pitch_width is not None:
            return NormalizedPitchDimensions(
                x_dim=Dimension(-1, 1),
                y_dim=Dimension(-1, 1),
                pitch_length=self._pitch_length,
                pitch_width=self._pitch_width,
                standardized=False,
            )
        else:
            return NormalizedPitchDimensions(
                x_dim=Dimension(-1, 1),
                y_dim=Dimension(-1, 1),
                pitch_length=105,
                pitch_width=68,
                standardized=True,
            )


class SportVUCoordinateSystem(ProviderCoordinateSystem):
    """
    StatsPerform SportVU coordinate system.

    Uses a pitch with the origin at the top left and the y-axis oriented
    from top to bottom. The coordinates are in meters.
    """

    @property
    def provider(self) -> Provider:
        return Provider.SPORTVU

    @property
    def origin(self) -> Origin:
        return Origin.TOP_LEFT

    @property
    def vertical_orientation(self) -> VerticalOrientation:
        return VerticalOrientation.TOP_TO_BOTTOM

    @property
    def pitch_dimensions(self) -> PitchDimensions:
        return MetricPitchDimensions(
            x_dim=Dimension(0, self._pitch_length),
            y_dim=Dimension(0, self._pitch_width),
            pitch_length=self._pitch_length,
            pitch_width=self._pitch_width,
            standardized=False,
        )


class HawkEyeCoordinateSystem(ProviderCoordinateSystem):
    """
    HawkEye coordinate system.

    Uses a pitch with the origin at the center and the y-axis oriented
    from bottom to top. The coordinates are in meters.
    """

    @property
    def provider(self) -> Provider:
        return Provider.HAWKEYE

    @property
    def origin(self) -> Origin:
        return Origin.CENTER

    @property
    def vertical_orientation(self) -> VerticalOrientation:
        return VerticalOrientation.BOTTOM_TO_TOP

    @property
    def pitch_dimensions(self) -> PitchDimensions:
        if self._pitch_length is not None and self._pitch_width is not None:
            return MetricPitchDimensions(
                x_dim=Dimension(
                    -1 * self._pitch_length / 2, self._pitch_length / 2
                ),
                y_dim=Dimension(
                    -1 * self._pitch_width / 2, self._pitch_width / 2
                ),
                pitch_length=self._pitch_length,
                pitch_width=self._pitch_width,
                standardized=False,
            )
        else:
            return MetricPitchDimensions(
                x_dim=Dimension(None, None),
                y_dim=Dimension(None, None),
                pitch_length=None,
                pitch_width=None,
                standardized=False,
            )


class DatasetType(Enum):
    """
    Dataset types.

    Attributes:
        TRACKING (DatasetType): A dataset containing tracking data.
        EVENT (DatasetType): A dataset containing event data.
        CODE (DatasetType): A dataset containing SportsCode annotations.
    """

    TRACKING = "TRACKING"
    EVENT = "EVENT"
    CODE = "CODE"

    def __repr__(self):
        return self.value


def build_coordinate_system(
    provider: Provider,
    dataset_type: DatasetType = DatasetType.EVENT,
    pitch_length: Optional[float] = None,
    pitch_width: Optional[float] = None,
) -> CoordinateSystem:
    """Build a coordinate system for a given provider and dataset type.

    Args:
        provider: The provider of the dataset.
        dataset_type: The type of the dataset.
        pitch_length: The real length of the pitch.
        pitch_width: The real width of the pitch.

    Returns:
        The coordinate system for the given provider and dataset type.
    """
    coordinate_systems = {
        Provider.TRACAB: TracabCoordinateSystem,
        Provider.KLOPPY: KloppyCoordinateSystem,
        Provider.METRICA: MetricaCoordinateSystem,
        Provider.OPTA: OptaCoordinateSystem,
        Provider.SPORTEC: {
            DatasetType.EVENT: SportecEventDataCoordinateSystem,
            DatasetType.TRACKING: SportecTrackingDataCoordinateSystem,
        },
        Provider.STATSBOMB: StatsBombCoordinateSystem,
        Provider.PFF: PFFCoordinateSystem,
        Provider.WYSCOUT: WyscoutCoordinateSystem,
        Provider.SKILLCORNER: SkillCornerCoordinateSystem,
        Provider.DATAFACTORY: DatafactoryCoordinateSystem,
        Provider.SECONDSPECTRUM: SecondSpectrumCoordinateSystem,
        Provider.HAWKEYE: HawkEyeCoordinateSystem,
        Provider.SPORTVU: SportVUCoordinateSystem,
        Provider.SIGNALITY: SignalityCoordinateSystem,
    }

    if provider in coordinate_systems:
        if isinstance(coordinate_systems[provider], dict):
            assert dataset_type in coordinate_systems[provider]
            return coordinate_systems[provider][dataset_type](
                pitch_length=pitch_length, pitch_width=pitch_width
            )
        else:
            return coordinate_systems[provider](
                pitch_length=pitch_length, pitch_width=pitch_width
            )
    else:
        raise ValueError(f"Invalid provider: {provider}")


class DatasetFlag(Flag):
    BALL_OWNING_TEAM = 1
    BALL_STATE = 2


@dataclass
class Statistic(ABC):
    name: str = field(init=False)


@dataclass
class ScalarStatistic(Statistic):
    value: float


@dataclass
class ExpectedGoals(ScalarStatistic):
    """Expected goals"""

    def __post_init__(self):
        self.name = "xG"


@dataclass
class PostShotExpectedGoals(ScalarStatistic):
    """Post-shot expected goals"""

    def __post_init__(self):
        self.name = "PSxG"


@dataclass
class ActionValue(Statistic):
    """Action value"""

    name: str
    action_value_scoring_before: Optional[float] = field(default=None)
    action_value_scoring_after: Optional[float] = field(default=None)
    action_value_conceding_before: Optional[float] = field(default=None)
    action_value_conceding_after: Optional[float] = field(default=None)

    @property
    def offensive_value(self) -> Optional[float]:
        if (
            self.action_value_scoring_before is None
            or self.action_value_scoring_after is None
        ):
            return None
        return (
            self.action_value_scoring_after - self.action_value_scoring_before
        )

    @property
    def defensive_value(self) -> Optional[float]:
        if (
            self.action_value_conceding_before is None
            or self.action_value_conceding_after is None
        ):
            return None
        return (
            self.action_value_conceding_after
            - self.action_value_conceding_before
        )

    @property
    def value(self) -> Optional[float]:
        if self.offensive_value is None or self.defensive_value is None:
            return None
        return self.offensive_value - self.defensive_value


@dataclass
class DataRecord(ABC):
    """
    Base class for a data record in a dataset.

    Attributes:
        dataset: The dataset to which the record belongs.
        prev_record: The previous record in the dataset.
        next_record: The next record in the dataset.
        record_id: The unique identifier of the record. Given by the provider.
        period: The match period in which the observation occurred.
        timestamp: Timestamp of occurrence, relative to the period kick-off.
        time: The time of the observation. Combines `period` and `timestamp`.
        attacking_direction: The attacking direction of the home team.
        ball_owning_team: The team that had possession of the ball.
        ball_state: The state of the ball at the time of the observation.
    """

    dataset: Dataset = field(init=False)
    prev_record: Optional[Self] = field(init=False)
    next_record: Optional[Self] = field(init=False)
    period: Period
    timestamp: timedelta
    statistics: List[Statistic]
    ball_owning_team: Optional[Team]
    ball_state: Optional[BallState]

    @property
    @abstractmethod
    def record_id(self) -> Union[int, str]:
        pass

    @property
    def time(self) -> Time:
        return Time(period=self.period, timestamp=self.timestamp)

    def set_refs(
        self,
        dataset: Dataset,
        prev: Optional[Self],
        next_: Optional[Self],
    ):
        if hasattr(self, "dataset"):
            # TODO: determine if next/prev record should be affected
            # by Dataset.filter
            return

        self.dataset = dataset
        self.prev_record = prev
        self.next_record = next_

    @property
    def attacking_direction(self):
        if (
            self.dataset
            and self.dataset.metadata
            and self.dataset.metadata.orientation is not None
        ):
            try:
                return AttackingDirection.from_orientation(
                    self.dataset.metadata.orientation,
                    period=self.period,
                    ball_owning_team=self.ball_owning_team,
                )
            except OrientationError:
                return AttackingDirection.NOT_SET
        return AttackingDirection.NOT_SET

    def matches(
        self, filter_: Optional[Union[str, Callable[[Self], bool]]]
    ) -> bool:
        if filter_ is None:
            return True
        elif callable(filter_):
            return filter_(self)
        else:
            raise InvalidFilterError()

    def prev(
        self, filter_: Optional[Union[str, Callable[[Self], bool]]] = None
    ) -> Optional[Self]:
        if self.prev_record:
            prev_record = self.prev_record
            while prev_record:
                if prev_record.matches(filter_):
                    return prev_record
                prev_record = prev_record.prev_record

    def next(
        self, filter_: Optional[Union[str, Callable[[Self], bool]]] = None
    ) -> Optional[Self]:
        if self.next_record:
            next_record = self.next_record
            while next_record:
                if next_record.matches(filter_):
                    return next_record
                next_record = next_record.next_record

    def replace(self, **changes):
        return replace(self, **changes)

    def __str__(self):
        return f"<{self.__class__.__name__}>"

    __repr__ = __str__


@dataclass
class Metadata:
    """
    Metadata for a dataset.

    Metadata is additional information about the dataset that is not part of
    the actual data. It includes information about the teams, the pitch
    dimensions, the orientation, the provider, and more.

    Attributes:
        game_id: Game id of the game from the provider.
        date: Date of the game.
        game_week: Game week (or match day) of the game. It can also be the stage
            (ex: "8th Finals"), if the game is happening during a cup or a play-off.
        periods: List of [`Period`][kloppy.domain.models.common.Period] entities.
        teams: `[home_team, away_team]`.
        coordinate_system: The coordinate system in which the data is provided.
        pitch_dimensions: The dimensions of the pitch.
        orientation: The attacking direction of each team.
        flags: Flags describing what optional data is available in the dataset.
        provider: The provider of the dataset.
        score: The final score of the match.
        frame_rate: The frame rate (in Hertz) at which the data was recorded.
            Only for tracking data.
        attributes: Additional metadata.
    """

    periods: List[Period]
    teams: List[Team]
    coordinate_system: CoordinateSystem
    pitch_dimensions: PitchDimensions
    orientation: Orientation
    flags: DatasetFlag
    provider: Provider
    score: Optional[Score] = None
    frame_rate: Optional[float] = None
    date: Optional[datetime] = None
    game_week: Optional[str] = None
    game_id: Optional[str] = None
    home_coach: Optional[str] = None
    away_coach: Optional[str] = None
    officials: Optional[List] = field(default_factory=list)
    attributes: Optional[Dict] = field(default_factory=dict, compare=False)

    def __post_init__(self):
        if self.coordinate_system is not None:
            # set the pitch dimensions from the coordinate system
            self.pitch_dimensions = self.coordinate_system.pitch_dimensions

        for i, period in enumerate(self.periods):
            period.set_refs(
                prev=self.periods[i - 1] if i > 0 else None,
                next_=(
                    self.periods[i + 1] if i + 1 < len(self.periods) else None
                ),
            )


T = TypeVar("T", bound="DataRecord")


@dataclass
class Dataset(ABC, Generic[T]):
    """
    Base class for datasets.

    A dataset describes specific aspects of what happened during a single
    match as a sequence of [`DataRecord`][kloppy.domain.DataRecord] entities.

    Attributes:
        dataset_type: The type of the dataset.
        records: List of records in the dataset.
        metadata: Metadata for the dataset.
    """

    records: List[T]
    metadata: Metadata

    def __iter__(self):
        return iter(self.records)

    def __getitem__(self, item):
        return self.records[item]

    def __len__(self):
        return len(self.records)

    def __post_init__(self):
        for i, record in enumerate(self.records):
            record.set_refs(
                dataset=self,
                prev=self.records[i - 1] if i > 0 else None,
                next_=(
                    self.records[i + 1] if i + 1 < len(self.records) else None
                ),
            )

        self._init_player_positions()
        self._update_formations_and_positions()

    def _init_player_positions(self):
        start_of_match = self.metadata.periods[0].start_time
        for team in self.metadata.teams:
            for player in team.players:
                player.positions.reset()
                if player.starting:
                    player.set_position(
                        start_of_match,
                        player.starting_position or PositionType.unknown(),
                    )

    def _update_formations_and_positions(self):
        """Update player positions based on the events for example."""
        pass

    @property
    @abstractmethod
    def dataset_type(self) -> DatasetType:
        raise NotImplementedError

    @abstractmethod
    def to_pandas(
        self,
        record_converter: Optional[Callable[[T], Dict]] = None,
        additional_columns: Optional[NamedColumns] = None,
    ) -> "DataFrame":
        pass

    def transform(self, *args, **kwargs):
        """
        See [transform][kloppy.helpers.transform]
        """
        from kloppy.helpers import transform

        return transform(self, *args, **kwargs)

    def filter(self, filter_: Union[str, Callable[[T], bool]]):
        """
        Filter all records used `filter_`

        Args:
            filter_: The filter to be used to filter the records. It can be a
                callable that takes a record and returns a boolean, or a string
                representing a css-like selector.

        Examples:

            >>> from kloppy.domain import EventType
            >>> dataset = dataset.filter(lambda event: event.event_type == EventType.PASS)
            >>> dataset = dataset.filter('pass')
        """
        return replace(
            self,
            records=self.find_all(filter_),
        )

    def map(self, mapper):
        return replace(
            self, records=[mapper(record) for record in self.records]
        )

    def find_all(self, filter_) -> List[T]:
        return [record for record in self.records if record.matches(filter_)]

    def find(self, filter_) -> Optional[T]:
        for record in self.records:
            if record.matches(filter_):
                return record

    @classmethod
    def from_dataset(cls, dataset: Dataset, mapper_fn: Callable[[Self], Self]):
        """
        Create a new Dataset from other dataset

        Arguments:
            mapper_fn:

        Examples:
            >>> from kloppy.domain import Code,     CodeDataset

            >>> code_dataset = (
            >>>     CodeDataset
            >>>     .from_dataset(
            >>>         dataset,
            >>>         lambda event: Code(
            >>>             code_id=event.event_id,
            >>>             code=event.event_name,
            >>>             period=event.period,
            >>>             timestamp=event.timestamp - 7,
            >>>             end_timestamp=event.timestamp + 5,
            >>>             labels={
            >>>                 'Player': str(event.player),
            >>>                 'Team': str(event.team)
            >>>             }
            >>>         )
            >>>     )
            >>> )
        """
        return cls(
            metadata=dataset.metadata,
            records=[mapper_fn(record) for record in dataset.records],
        )

    def get_record_by_id(self, record_id: Union[int, str]) -> Optional[T]:
        for record in self.records:
            if record.record_id == record_id:
                return record

    @overload
    def to_records(
        self,
        *columns: Unpack[tuple[Column]],
        as_list: Literal[True] = True,
        **named_columns: NamedColumns,
    ) -> List[Dict[str, Any]]:
        ...

    @overload
    def to_records(
        self,
        *columns: Unpack[tuple[Column]],
        as_list: Literal[False] = False,
        **named_columns: NamedColumns,
    ) -> Iterable[Dict[str, Any]]:
        ...

    def to_records(
        self,
        *columns: Unpack[tuple[Column]],
        as_list: bool = True,
        **named_columns: NamedColumns,
    ) -> Union[List[Dict[str, Any]], Iterable[Dict[str, Any]]]:
        from ..services.transformers.data_record import get_transformer_cls

        transformer = get_transformer_cls(self.dataset_type)(
            *columns, **named_columns
        )
        iterator = map(transformer, self.records)
        if as_list:
            return list(iterator)
        else:
            return iterator

    def to_dict(
        self,
        *columns: Unpack[tuple[Column]],
        orient: Literal["list"] = "list",
        **named_columns: NamedColumns,
    ) -> Dict[str, List[Any]]:
        if orient == "list":
            from ..services.transformers.data_record import get_transformer_cls

            transformer = get_transformer_cls(self.dataset_type)(
                *columns, **named_columns
            )

            c = len(self.records)
            items = defaultdict(lambda: [None] * c)
            for i, record in enumerate(self.records):
                item = transformer(record)
                for k, v in item.items():
                    items[k][i] = v

            return items
        else:
            raise KloppyParameterError(
                f"Orient {orient} is not supported. Only orient='list' is supported"
            )

    def to_df(
        self,
        *columns: Unpack[tuple[Column]],
        engine: Optional[
            Union[
                Literal["polars"],
                Literal["pandas"],
                Literal["pandas[pyarrow]"],
            ]
        ] = None,
        **named_columns: NamedColumns,
    ):
        from kloppy.config import get_config

        if not engine:
            engine = get_config("dataframe.engine")

        if engine == "pandas[pyarrow]":
            try:
                import pandas as pd

                types_mapper = pd.ArrowDtype
            except ImportError:
                raise ImportError(
                    "Seems like you don't have pandas installed. Please"
                    " install it using: pip install pandas"
                )
            except AttributeError:
                raise AttributeError(
                    "Seems like you have an older version of pandas installed. Please"
                    " upgrade to at least 1.5 using: pip install pandas>=1.5"
                )

            try:
                import pyarrow as pa
            except ImportError:
                raise ImportError(
                    "Seems like you don't have pyarrow installed. Please"
                    " install it using: pip install pyarrow"
                )

            table = pa.Table.from_pydict(
                self.to_dict(*columns, orient="list", **named_columns)
            )
            return table.to_pandas(types_mapper=types_mapper)

        elif engine == "pandas":
            try:
                from pandas import DataFrame
            except ImportError:
                raise ImportError(
                    "Seems like you don't have pandas installed. Please"
                    " install it using: pip install pandas"
                )

            return DataFrame.from_dict(
                self.to_dict(*columns, orient="list", **named_columns)
            )
        elif engine == "polars":
            try:
                from polars import from_dict
            except ImportError:
                raise ImportError(
                    "Seems like you don't have polars installed. Please"
                    " install it using: pip install polars"
                )

            return from_dict(
                self.to_dict(*columns, orient="list", **named_columns)
            )
        else:
            raise KloppyParameterError(f"Engine {engine} is not valid")

    def __repr__(self):
        return f"<{self.__class__.__name__} record_count={len(self.records)}>"

    __str__ = __repr__



from typing import List, Optional

from kloppy.config import get_config
from kloppy.domain import EventDataset, EventFactory, Provider, TrackingDataset
from kloppy.infra.serializers.event.statsperform import (
    StatsPerformDeserializer as StatsPerformEventDeserializer,
)
from kloppy.infra.serializers.event.statsperform import (
    StatsPerformInputs as StatsPerformEventInputs,
)
from kloppy.infra.serializers.tracking.statsperform import (
    StatsPerformDeserializer as StatsPerformTrackingDeserializer,
)
from kloppy.infra.serializers.tracking.statsperform import (
    StatsPerformInputs as StatsPerformTrackingInputs,
)
from kloppy.io import FileLike, open_as_file
from kloppy.utils import deprecated


@deprecated("statsperform.load_tracking should be used")
def load(
    meta_data: FileLike,  # Stats Perform MA1 file - xml or json - single game, live data & lineups
    raw_data: FileLike,  # Stats Perform MA25 file - txt - tracking data
    tracking_system: str = "sportvu",
    pitch_length: Optional[float] = None,
    pitch_width: Optional[float] = None,
    sample_rate: Optional[float] = None,
    limit: Optional[int] = None,
    coordinates: Optional[str] = None,
    only_alive: Optional[bool] = False,
) -> TrackingDataset:
    deserializer = StatsPerformTrackingDeserializer(
        provider=Provider[tracking_system.upper()],
        sample_rate=sample_rate,
        limit=limit,
        coordinate_system=coordinates,
        only_alive=only_alive,
    )
    with open_as_file(meta_data) as meta_data_fp, open_as_file(
        raw_data
    ) as raw_data_fp:
        return deserializer.deserialize(
            inputs=StatsPerformTrackingInputs(
                meta_data=meta_data_fp,
                raw_data=raw_data_fp,
                pitch_length=pitch_length,
                pitch_width=pitch_width,
            )
        )


def load_event(
    ma1_data: FileLike,
    ma3_data: FileLike,
    pitch_length: Optional[float] = None,
    pitch_width: Optional[float] = None,
    event_types: Optional[List[str]] = None,
    coordinates: Optional[str] = None,
    event_factory: Optional[EventFactory] = None,
) -> EventDataset:
    """Load Stats Perform event data.

    Args:
        ma1_data: A MA1 json or xml feed containing the lineup information.
        ma3_data: A MA3 json or xml feed containing the events.
        pitch_length: The length of the pitch (in meters).
        pitch_width: The width of the pitch (in meters).
        event_types: A list of event types to load.
        coordinates: The coordinate system to use.
        event_factory: A custom event factory.

    Returns:
        The parsed event data.
    """
    deserializer = StatsPerformEventDeserializer(
        event_types=event_types,
        coordinate_system=coordinates,
        event_factory=event_factory or get_config("event_factory"),  # type: ignore
    )
    with open_as_file(ma1_data) as ma1_data_fp, open_as_file(
        ma3_data
    ) as ma3_data_fp:
        return deserializer.deserialize(
            inputs=StatsPerformEventInputs(
                meta_data=ma1_data_fp,
                meta_feed="MA1",
                event_data=ma3_data_fp,
                event_feed="MA3",
                pitch_length=pitch_length,
                pitch_width=pitch_width,
            ),
        )


def load_tracking(
    ma1_data: FileLike,
    ma25_data: FileLike,
    tracking_system: str = "sportvu",
    pitch_length: Optional[float] = None,
    pitch_width: Optional[float] = None,
    sample_rate: Optional[float] = None,
    limit: Optional[int] = None,
    coordinates: Optional[str] = None,
    only_alive: Optional[bool] = False,
) -> TrackingDataset:
    """
    Load Stats Perform tracking data.

    Args:
        ma1_data: A json or xml feed containing the lineup information.
        ma25_data: A txt file linked in the MA25 Match Tracking Feed; also known as an OPT file.
        tracking_system: The system that generated the tracking data.
        pitch_length: The length of the pitch (in meters).
        pitch_width: The width of the pitch (in meters).
        sample_rate: Sample the data at a specific rate.
        limit: Limit the number of frames to load to the first `limit` frames.
        coordinates: The coordinate system to use.
        only_alive: Only include frames in which the game is not paused.

    Returns:
        The parsed tracking data.
    """
    deserializer = StatsPerformTrackingDeserializer(
        provider=Provider[tracking_system.upper()],
        sample_rate=sample_rate,
        limit=limit,
        coordinate_system=coordinates,
        only_alive=only_alive,
    )
    with open_as_file(ma1_data) as ma1_data_fp, open_as_file(
        ma25_data
    ) as ma25_data_fp:
        return deserializer.deserialize(
            inputs=StatsPerformTrackingInputs(
                meta_data=ma1_data_fp,
                raw_data=ma25_data_fp,
                pitch_length=pitch_length,
                pitch_width=pitch_width,
            )
        )



import logging
from collections import OrderedDict
from datetime import datetime, timedelta
from typing import IO, Dict, List, NamedTuple

from lxml import objectify

from kloppy.domain import (
    BallState,
    BodyPart,
    BodyPartQualifier,
    CardType,
    DatasetFlag,
    EventDataset,
    EventType,
    Ground,
    Metadata,
    Official,
    OfficialType,
    Orientation,
    PassResult,
    Period,
    Player,
    Point,
    PositionType,
    Provider,
    Qualifier,
    Score,
    SetPieceQualifier,
    SetPieceType,
    ShotResult,
    Team,
)
from kloppy.exceptions import DeserializationError
from kloppy.infra.serializers.event.deserializer import EventDataDeserializer
from kloppy.utils import performance_logging

position_types_mapping: Dict[str, PositionType] = {
    "TW": PositionType.Goalkeeper,
    "IVR": PositionType.RightCenterBack,
    "IVL": PositionType.LeftCenterBack,
    "STR": PositionType.Striker,
    "STL": PositionType.LeftForward,
    "STZ": PositionType.Striker,
    "ZO": PositionType.CenterAttackingMidfield,
    "LV": PositionType.LeftBack,
    "RV": PositionType.RightBack,
    "DMR": PositionType.RightDefensiveMidfield,
    "DRM": PositionType.RightDefensiveMidfield,
    "DML": PositionType.LeftDefensiveMidfield,
    "DLM": PositionType.LeftDefensiveMidfield,
    "ORM": PositionType.RightMidfield,
    "OLM": PositionType.LeftMidfield,
    "RA": PositionType.RightWing,
    "LA": PositionType.LeftWing,
}

referee_types_mapping: Dict[str, OfficialType] = {
    "referee": OfficialType.MainReferee,
    "firstAssistant": OfficialType.AssistantReferee,
    "videoReferee": OfficialType.VideoAssistantReferee,
    "videoRefereeAssistant": OfficialType.AssistantVideoAssistantReferee,
    "secondAssistant": OfficialType.AssistantReferee,
    "fourthOfficial": OfficialType.FourthOfficial,
}

logger = logging.getLogger(__name__)


def _team_from_xml_elm(team_elm) -> Team:
    team = Team(
        team_id=team_elm.attrib["TeamId"],
        name=team_elm.attrib["TeamName"],
        ground=(
            Ground.HOME if team_elm.attrib["Role"] == "home" else Ground.AWAY
        ),
    )
    team.players = [
        Player(
            player_id=player_elm.attrib["PersonId"],
            team=team,
            jersey_no=int(player_elm.attrib["ShirtNumber"]),
            name=player_elm.attrib["Shortname"],
            first_name=player_elm.attrib["FirstName"],
            last_name=player_elm.attrib["LastName"],
            starting_position=position_types_mapping.get(
                player_elm.attrib.get("PlayingPosition"), PositionType.Unknown
            ),
            starting=player_elm.attrib["Starting"] == "true",
        )
        for player_elm in team_elm.Players.iterchildren("Player")
    ]
    return team


SPORTEC_FPS = 25

"""Sportec uses fixed starting frame ids for each half"""
SPORTEC_FIRST_HALF_STARTING_FRAME_ID = 10_000
SPORTEC_SECOND_HALF_STARTING_FRAME_ID = 100_000
SPORTEC_FIRST_EXTRA_HALF_STARTING_FRAME_ID = 200_000
SPORTEC_SECOND_EXTRA_HALF_STARTING_FRAME_ID = 250_000


class SportecMetadata(NamedTuple):
    score: Score
    teams: List[Team]
    periods: List[Period]
    x_max: float
    y_max: float
    fps: int
    home_coach: str
    away_coach: str
    officials: List[Official]


def sportec_metadata_from_xml_elm(match_root) -> SportecMetadata:
    """
    Load metadata from Sportec XML element. This part is shared between event- and tracking data.
    In the future this might move to a common.sportec package that provides functionality for both
    deserializers.
    """
    x_max = float(match_root.MatchInformation.Environment.attrib["PitchX"])
    y_max = float(match_root.MatchInformation.Environment.attrib["PitchY"])

    team_path = objectify.ObjectPath("PutDataRequest.MatchInformation.Teams")
    team_elms = list(team_path.find(match_root).iterchildren("Team"))

    home_team = away_team = None
    for team_elm in team_elms:
        head_coach = [
            trainer.attrib["Shortname"]
            for trainer in team_elm.TrainerStaff.iterchildren("Trainer")
            if trainer.attrib["Role"] == "headcoach"
        ]
        if team_elm.attrib["Role"] == "home":
            home_team = _team_from_xml_elm(team_elm)
            home_coach = head_coach[0] if len(head_coach) else None
        elif team_elm.attrib["Role"] == "guest":
            away_team = _team_from_xml_elm(team_elm)
            away_coach = head_coach[0] if len(head_coach) else None
        else:
            raise DeserializationError(
                f"Unknown side: {team_elm.attrib['Role']}"
            )

    if not home_team:
        raise DeserializationError("Home team is missing from metadata")
    if not away_team:
        raise DeserializationError("Away team is missing from metadata")

    (home_score, away_score,) = match_root.MatchInformation.General.attrib[
        "Result"
    ].split(":")
    score = Score(home=int(home_score), away=int(away_score))
    teams = [home_team, away_team]

    if len(home_team.players) == 0 or len(away_team.players) == 0:
        raise DeserializationError("LineUp incomplete")

    # The periods can be rebuild from event data. Therefore, the periods attribute
    # from the metadata can be ignored. It is required for tracking data.
    other_game_information = (
        match_root.MatchInformation.OtherGameInformation.attrib
    )
    periods = [
        Period(
            id=1,
            start_timestamp=timedelta(
                seconds=SPORTEC_FIRST_HALF_STARTING_FRAME_ID / SPORTEC_FPS
            ),
            end_timestamp=timedelta(
                seconds=SPORTEC_FIRST_HALF_STARTING_FRAME_ID / SPORTEC_FPS
                + float(other_game_information["TotalTimeFirstHalf"]) / 1000
            ),
        ),
        Period(
            id=2,
            start_timestamp=timedelta(
                seconds=SPORTEC_SECOND_HALF_STARTING_FRAME_ID / SPORTEC_FPS
            ),
            end_timestamp=timedelta(
                seconds=SPORTEC_SECOND_HALF_STARTING_FRAME_ID / SPORTEC_FPS
                + float(other_game_information["TotalTimeSecondHalf"]) / 1000
            ),
        ),
    ]

    if "TotalTimeFirstHalfExtra" in other_game_information:
        # Add two periods for extra time.
        periods.extend(
            [
                Period(
                    id=3,
                    start_timestamp=timedelta(
                        seconds=SPORTEC_FIRST_EXTRA_HALF_STARTING_FRAME_ID
                        / SPORTEC_FPS
                    ),
                    end_timestamp=timedelta(
                        seconds=SPORTEC_FIRST_EXTRA_HALF_STARTING_FRAME_ID
                        / SPORTEC_FPS
                        + float(
                            other_game_information["TotalTimeFirstHalfExtra"]
                        )
                        / 1000
                    ),
                ),
                Period(
                    id=4,
                    start_timestamp=timedelta(
                        seconds=SPORTEC_SECOND_EXTRA_HALF_STARTING_FRAME_ID
                        / SPORTEC_FPS
                    ),
                    end_timestamp=timedelta(
                        seconds=SPORTEC_SECOND_EXTRA_HALF_STARTING_FRAME_ID
                        / SPORTEC_FPS
                        + float(
                            other_game_information["TotalTimeSecondHalfExtra"]
                        )
                        / 1000
                    ),
                ),
            ]
        )

    if hasattr(match_root, "MatchInformation") and hasattr(
        match_root.MatchInformation, "Referees"
    ):
        officials = []
        referee_path = objectify.ObjectPath(
            "PutDataRequest.MatchInformation.Referees"
        )
        referee_elms = referee_path.find(match_root).iterchildren(
            tag="Referee"
        )

        for referee in referee_elms:
            ref_attrib = referee.attrib
            officials.append(
                Official(
                    official_id=ref_attrib["PersonId"],
                    name=ref_attrib["Shortname"],
                    first_name=ref_attrib["FirstName"],
                    last_name=ref_attrib["LastName"],
                    role=referee_types_mapping.get(
                        ref_attrib["Role"], OfficialType.Unknown
                    ),
                )
            )
    else:
        officials = []

    return SportecMetadata(
        score=score,
        teams=teams,
        periods=periods,
        x_max=x_max,
        y_max=y_max,
        fps=SPORTEC_FPS,
        home_coach=home_coach,
        away_coach=away_coach,
        officials=officials,
    )


def _event_chain_from_xml_elm(event_elm):
    chain = OrderedDict()
    current_elm = event_elm
    while True:
        chain[current_elm.tag] = dict(current_elm.attrib)
        if not current_elm.countchildren():
            break
        current_elm = current_elm.getchildren()[0]
    return chain


SPORTEC_EVENT_NAME_KICKOFF = "KickOff"
SPORTEC_EVENT_NAME_FINAL_WHISTLE = "FinalWhistle"

SPORTEC_EVENT_NAME_SHOT_WIDE = "ShotWide"
SPORTEC_EVENT_NAME_SHOT_SAVED = "SavedShot"
SPORTEC_EVENT_NAME_SHOT_BLOCKED = "BlockedShot"
SPORTEC_EVENT_NAME_SHOT_WOODWORK = "ShotWoodWork"
SPORTEC_EVENT_NAME_SHOT_OTHER = "OtherShot"
SPORTEC_EVENT_NAME_SHOT_GOAL = "SuccessfulShot"
SPORTEC_EVENT_NAME_OWN_GOAL = "OwnGoal"
SPORTEC_SHOT_EVENT_NAMES = (
    SPORTEC_EVENT_NAME_SHOT_WIDE,
    SPORTEC_EVENT_NAME_SHOT_SAVED,
    SPORTEC_EVENT_NAME_SHOT_BLOCKED,
    SPORTEC_EVENT_NAME_SHOT_WOODWORK,
    SPORTEC_EVENT_NAME_SHOT_OTHER,
    SPORTEC_EVENT_NAME_SHOT_GOAL,
    SPORTEC_EVENT_NAME_OWN_GOAL,
)

SPORTEC_EVENT_NAME_PASS = "Pass"
SPORTEC_EVENT_NAME_CROSS = "Cross"
SPORTEC_EVENT_NAME_THROW_IN = "ThrowIn"
SPORTEC_EVENT_NAME_GOAL_KICK = "GoalKick"
SPORTEC_EVENT_NAME_PENALTY = "Penalty"
SPORTEC_EVENT_NAME_CORNER_KICK = "CornerKick"
SPORTEC_EVENT_NAME_FREE_KICK = "FreeKick"
SPORTEC_PASS_EVENT_NAMES = (SPORTEC_EVENT_NAME_PASS, SPORTEC_EVENT_NAME_CROSS)

SPORTEC_EVENT_NAME_BALL_CLAIMING = "BallClaiming"
SPORTEC_EVENT_NAME_SUBSTITUTION = "Substitution"
SPORTEC_EVENT_NAME_CAUTION = "Caution"
SPORTEC_EVENT_NAME_FOUL = "Foul"

SPORTEC_EVENT_TYPE_OF_SHOT = "TypeOfShot"
SPORTEC_EVENT_BODY_PART_HEAD = "head"
SPORTEC_EVENT_BODY_PART_LEFT_FOOT = "leftLeg"
SPORTEC_EVENT_BODY_PART_RIGHT_FOOT = "rightLeg"


def _parse_datetime(dt_str: str) -> datetime:
    return datetime.fromisoformat(dt_str)


def _get_event_qualifiers(event_chain: Dict) -> List[Qualifier]:
    qualifiers = []

    qualifiers.extend(_get_event_setpiece_qualifiers(event_chain))
    qualifiers.extend(_get_event_bodypart_qualifiers(event_chain))

    return qualifiers


def _get_event_setpiece_qualifiers(event_chain):
    qualifiers = []

    if SPORTEC_EVENT_NAME_THROW_IN in event_chain:
        qualifiers.append(SetPieceQualifier(value=SetPieceType.THROW_IN))
    elif SPORTEC_EVENT_NAME_GOAL_KICK in event_chain:
        qualifiers.append(SetPieceQualifier(value=SetPieceType.GOAL_KICK))
    elif SPORTEC_EVENT_NAME_PENALTY in event_chain:
        qualifiers.append(SetPieceQualifier(value=SetPieceType.PENALTY))
    elif SPORTEC_EVENT_NAME_CORNER_KICK in event_chain:
        qualifiers.append(SetPieceQualifier(value=SetPieceType.CORNER_KICK))
    elif SPORTEC_EVENT_NAME_KICKOFF in event_chain:
        qualifiers.append(SetPieceQualifier(value=SetPieceType.KICK_OFF))
    elif SPORTEC_EVENT_NAME_FREE_KICK in event_chain:
        qualifiers.append(SetPieceQualifier(value=SetPieceType.FREE_KICK))

    return qualifiers


def _get_event_bodypart_qualifiers(event_chain):
    qualifiers = []

    if SPORTEC_EVENT_BODY_PART_HEAD in [
        item.get(SPORTEC_EVENT_TYPE_OF_SHOT) for item in event_chain.values()
    ]:
        qualifiers.append(BodyPartQualifier(value=BodyPart.HEAD))
    elif SPORTEC_EVENT_BODY_PART_LEFT_FOOT in [
        item.get(SPORTEC_EVENT_TYPE_OF_SHOT) for item in event_chain.values()
    ]:
        qualifiers.append(BodyPartQualifier(value=BodyPart.LEFT_FOOT))
    elif SPORTEC_EVENT_BODY_PART_RIGHT_FOOT in [
        item.get(SPORTEC_EVENT_TYPE_OF_SHOT) for item in event_chain.values()
    ]:
        qualifiers.append(BodyPartQualifier(value=BodyPart.RIGHT_FOOT))

    return qualifiers


def _parse_shot(event_name: str, event_chain: OrderedDict) -> Dict:
    if event_name == SPORTEC_EVENT_NAME_SHOT_WIDE:
        result = ShotResult.OFF_TARGET
    elif event_name == SPORTEC_EVENT_NAME_SHOT_SAVED:
        result = ShotResult.SAVED
    elif event_name == SPORTEC_EVENT_NAME_SHOT_BLOCKED:
        result = ShotResult.BLOCKED
    elif event_name == SPORTEC_EVENT_NAME_SHOT_WOODWORK:
        result = ShotResult.POST
    elif event_name == SPORTEC_EVENT_NAME_SHOT_GOAL:
        result = ShotResult.GOAL
    elif event_name == SPORTEC_EVENT_NAME_OWN_GOAL:
        result = ShotResult.OWN_GOAL
    elif event_name == SPORTEC_EVENT_NAME_SHOT_OTHER:
        result = None
    else:
        raise ValueError(f"Unknown shot type {event_name}")

    return dict(result=result, qualifiers=_get_event_qualifiers(event_chain))


def _parse_pass(event_chain: OrderedDict, team: Team) -> Dict:
    if event_chain["Play"]["Evaluation"] in (
        "successfullyCompleted",
        "successful",
    ):
        result = PassResult.COMPLETE
        if "Recipient" in event_chain["Play"]:
            receiver_player = team.get_player_by_id(
                event_chain["Play"]["Recipient"]
            )
        else:
            # this attribute can be missing according to docs
            receiver_player = None
    else:
        result = PassResult.INCOMPLETE
        receiver_player = None

    return dict(
        result=result,
        receiver_player=receiver_player,
        qualifiers=_get_event_qualifiers(event_chain),
    )


def _parse_substitution(event_attributes: Dict, team: Team) -> Dict:
    return dict(
        player=team.get_player_by_id(event_attributes["PlayerOut"]),
        replacement_player=team.get_player_by_id(event_attributes["PlayerIn"]),
    )


def _parse_caution(event_attributes: Dict) -> Dict:
    if event_attributes["CardColor"] == "yellow":
        card_type = CardType.FIRST_YELLOW
    elif event_attributes["CardColor"] == "yellowRed":
        card_type = CardType.SECOND_YELLOW
    elif event_attributes["CardColor"] == "red":
        card_type = CardType.RED
    else:
        raise ValueError(
            f"Unknown card color: {event_attributes['CardColor']}"
        )

    return dict(card_type=card_type)


def _parse_foul(event_attributes: Dict, teams: List[Team]) -> Dict:
    team = (
        teams[0]
        if event_attributes["TeamFouler"] == teams[0].team_id
        else teams[1]
    )
    player = team.get_player_by_id(event_attributes["Fouler"])

    return dict(team=team, player=player)


def _parse_coordinates(event_attributes: Dict) -> Point:
    if "X-Position" not in event_attributes:
        return None
    return Point(
        x=float(event_attributes["X-Position"]),
        y=float(event_attributes["Y-Position"]),
    )


class SportecEventDataInputs(NamedTuple):
    meta_data: IO[bytes]
    event_data: IO[bytes]


class SportecEventDataDeserializer(
    EventDataDeserializer[SportecEventDataInputs]
):
    @property
    def provider(self) -> Provider:
        return Provider.SPORTEC

    def deserialize(self, inputs: SportecEventDataInputs) -> EventDataset:
        with performance_logging("load data", logger=logger):
            match_root = objectify.fromstring(inputs.meta_data.read())
            event_root = objectify.fromstring(inputs.event_data.read())

        with performance_logging("parse data", logger=logger):
            date = datetime.fromisoformat(
                match_root.MatchInformation.General.attrib["KickoffTime"]
            )
            game_week = match_root.MatchInformation.General.attrib["MatchDay"]
            game_id = match_root.MatchInformation.General.attrib["MatchId"]

            sportec_metadata = sportec_metadata_from_xml_elm(match_root)
            teams = home_team, away_team = sportec_metadata.teams
            transformer = self.get_transformer(
                pitch_length=sportec_metadata.x_max,
                pitch_width=sportec_metadata.y_max,
            )
            home_coach = sportec_metadata.home_coach
            away_coach = sportec_metadata.away_coach

            periods = []
            period_id = 0
            events = []

            for event_elm in event_root.iterchildren("Event"):
                event_chain = _event_chain_from_xml_elm(event_elm)
                timestamp = _parse_datetime(event_chain["Event"]["EventTime"])

                if (
                    SPORTEC_EVENT_NAME_KICKOFF in event_chain
                    and "GameSection"
                    in event_chain[SPORTEC_EVENT_NAME_KICKOFF]
                ):
                    period_id += 1
                    period = Period(
                        id=period_id,
                        start_timestamp=timestamp,
                        end_timestamp=None,
                    )
                    if period_id == 1:
                        team_left = event_chain[SPORTEC_EVENT_NAME_KICKOFF][
                            "TeamLeft"
                        ]
                        orientation = (
                            Orientation.HOME_AWAY
                            if team_left == home_team.team_id
                            else Orientation.AWAY_HOME
                        )

                    periods.append(period)
                elif SPORTEC_EVENT_NAME_FINAL_WHISTLE in event_chain:
                    period.end_timestamp = timestamp
                    continue
                elif period_id == 0:
                    # Skip any events that happened before the first kick off
                    continue

                team = None
                player = None
                flatten_attributes = dict()
                # reverse because top levels are more important
                for event_attributes in reversed(event_chain.values()):
                    flatten_attributes.update(event_attributes)

                if "Team" in flatten_attributes:
                    team = (
                        home_team
                        if flatten_attributes["Team"] == home_team.team_id
                        else away_team
                    )
                if "Player" in flatten_attributes:
                    if not team:
                        raise ValueError("Player set while team is not set")
                    player = team.get_player_by_id(
                        flatten_attributes["Player"]
                    )

                generic_event_kwargs = dict(
                    # from DataRecord
                    period=period,
                    timestamp=timestamp - period.start_timestamp,
                    ball_owning_team=None,
                    ball_state=BallState.ALIVE,
                    # from Event
                    event_id=event_chain["Event"]["EventId"],
                    coordinates=_parse_coordinates(event_chain["Event"]),
                    raw_event=flatten_attributes,
                    team=team,
                    player=player,
                )

                event_name, event_attributes = event_chain.popitem()
                if event_name in SPORTEC_SHOT_EVENT_NAMES:
                    shot_event_kwargs = _parse_shot(
                        event_name=event_name, event_chain=event_chain
                    )
                    event = self.event_factory.build_shot(
                        **shot_event_kwargs,
                        **generic_event_kwargs,
                    )
                elif event_name in SPORTEC_PASS_EVENT_NAMES:
                    pass_event_kwargs = _parse_pass(
                        event_chain=event_chain, team=team
                    )
                    event = self.event_factory.build_pass(
                        **pass_event_kwargs,
                        **generic_event_kwargs,
                        receive_timestamp=None,
                        receiver_coordinates=None,
                    )
                elif event_name == SPORTEC_EVENT_NAME_BALL_CLAIMING:
                    event = self.event_factory.build_recovery(
                        result=None,
                        qualifiers=None,
                        **generic_event_kwargs,
                    )
                elif event_name == SPORTEC_EVENT_NAME_SUBSTITUTION:
                    substitution_event_kwargs = _parse_substitution(
                        event_attributes=event_attributes, team=team
                    )
                    generic_event_kwargs["player"] = substitution_event_kwargs[
                        "player"
                    ]
                    del substitution_event_kwargs["player"]
                    event = self.event_factory.build_substitution(
                        result=None,
                        qualifiers=None,
                        **substitution_event_kwargs,
                        **generic_event_kwargs,
                    )
                elif event_name == SPORTEC_EVENT_NAME_CAUTION:
                    card_kwargs = _parse_caution(event_attributes)
                    event = self.event_factory.build_card(
                        result=None,
                        qualifiers=None,
                        **card_kwargs,
                        **generic_event_kwargs,
                    )
                elif event_name == SPORTEC_EVENT_NAME_FOUL:
                    foul_kwargs = _parse_foul(event_attributes, teams=teams)
                    generic_event_kwargs.update(foul_kwargs)
                    event = self.event_factory.build_foul_committed(
                        result=None,
                        qualifiers=None,
                        **generic_event_kwargs,
                    )
                else:
                    event = self.event_factory.build_generic(
                        result=None,
                        qualifiers=None,
                        event_name=event_name,
                        **generic_event_kwargs,
                    )

                if (
                    event.event_type == EventType.PASS
                    and event.get_qualifier_value(SetPieceQualifier)
                    in (
                        SetPieceType.THROW_IN,
                        SetPieceType.GOAL_KICK,
                        SetPieceType.CORNER_KICK,
                    )
                ):
                    # 1. update previous pass
                    if events[-1].event_type == EventType.PASS:
                        events[-1].result = PassResult.OUT

                    # 2. add synthetic out event
                    decision_timestamp = _parse_datetime(
                        event_chain[list(event_chain.keys())[1]][
                            "DecisionTimestamp"
                        ]
                    )
                    out_event = self.event_factory.build_ball_out(
                        period=period,
                        timestamp=decision_timestamp - period.start_timestamp,
                        ball_owning_team=None,
                        ball_state=BallState.DEAD,
                        # from Event
                        event_id=event_chain["Event"]["EventId"] + "-ball-out",
                        team=events[-1].team,
                        player=events[-1].player,
                        coordinates=None,
                        raw_event={},
                        result=None,
                        qualifiers=None,
                    )
                    events.append(transformer.transform_event(out_event))

                events.append(transformer.transform_event(event))

        for i, event in enumerate(events[:-1]):
            if (
                event.event_type == EventType.PASS
                and event.result == PassResult.COMPLETE
            ):
                # Sportec uses X/Y-Source-Position to define the start coordinates of
                # an event and X/Y-Position to define the end of an event. There can/will
                # be quite a distance between the start and the end of an event.
                # When we want to set the receiver_coordinates we need to use
                # the start of the event.
                # How to solve this:
                # 1. Create a copy of an event
                # 2. Set the coordinates based on X/Y-Source-Position
                # 3. Pass through the transformer
                # 4. Update the receiver coordinates
                if "X-Source-Position" in events[i + 1].raw_event:
                    updated_event = transformer.transform_event(
                        events[i + 1].replace(
                            coordinates=Point(
                                x=float(
                                    events[i + 1].raw_event[
                                        "X-Source-Position"
                                    ]
                                ),
                                y=float(
                                    events[i + 1].raw_event[
                                        "Y-Source-Position"
                                    ]
                                ),
                            )
                        )
                    )
                    event.receiver_coordinates = updated_event.coordinates
                else:
                    event.receiver_coordinates = events[i + 1].coordinates

        events = list(
            filter(
                self.should_include_event,
                events,
            )
        )

        metadata = Metadata(
            teams=teams,
            periods=periods,
            pitch_dimensions=transformer.get_to_coordinate_system().pitch_dimensions,
            score=sportec_metadata.score,
            frame_rate=None,
            orientation=orientation,
            flags=~(DatasetFlag.BALL_STATE | DatasetFlag.BALL_OWNING_TEAM),
            provider=Provider.SPORTEC,
            coordinate_system=transformer.get_to_coordinate_system(),
            date=date,
            game_week=game_week,
            game_id=game_id,
            home_coach=home_coach,
            away_coach=away_coach,
            officials=sportec_metadata.officials,
        )

        return EventDataset(
            metadata=metadata,
            records=events,
        )



from enum import Enum


class PositionType(Enum):
    Unknown = ("Unknown", "UNK", None)

    Goalkeeper = ("Goalkeeper", "GK", None)

    Defender = ("Defender", "DEF", None)
    FullBack = ("Full Back", "FB", "Defender")
    LeftBack = ("Left Back", "LB", "FullBack")
    RightBack = ("Right Back", "RB", "FullBack")
    CenterBack = ("Center Back", "CB", "Defender")
    LeftCenterBack = ("Left Center Back", "LCB", "CenterBack")
    RightCenterBack = ("Right Center Back", "RCB", "CenterBack")
    LeftWingBack = ("Left Wing Back", "LWB", "WingBack")
    RightWingBack = ("Right Wing Back", "RWB", "WingBack")

    Midfielder = ("Midfielder", "MID", None)
    DefensiveMidfield = ("Defensive Midfield", "DM", "Midfielder")
    LeftDefensiveMidfield = (
        "Left Defensive Midfield",
        "LDM",
        "DefensiveMidfield",
    )
    CenterDefensiveMidfield = (
        "Center Defensive Midfield",
        "CDM",
        "DefensiveMidfield",
    )
    RightDefensiveMidfield = (
        "Right Defensive Midfield",
        "RDM",
        "DefensiveMidfield",
    )

    CentralMidfield = ("Central Midfield", "CM", "Midfielder")
    LeftCentralMidfield = ("Left Central Midfield", "LCM", "CentralMidfield")
    CenterMidfield = ("Center Midfield", "CM", "CentralMidfield")
    RightCentralMidfield = ("Right Central Midfield", "RCM", "CentralMidfield")

    AttackingMidfield = ("Attacking Midfield", "AM", "Midfielder")
    LeftAttackingMidfield = (
        "Left Attacking Midfield",
        "LAM",
        "AttackingMidfield",
    )
    CenterAttackingMidfield = (
        "Center Attacking Midfield",
        "CAM",
        "AttackingMidfield",
    )
    RightAttackingMidfield = (
        "Right Attacking Midfield",
        "RAM",
        "AttackingMidfield",
    )

    WideMidfield = ("Wide Midfield", "WM", "Midfielder")
    LeftWing = ("Left Wing", "LW", "WideMidfield")
    RightWing = ("Right Wing", "RW", "WideMidfield")
    LeftMidfield = ("Left Midfield", "LM", "WideMidfield")
    RightMidfield = ("Right Midfield", "RM", "WideMidfield")

    Attacker = ("Attacker", "ATT", None)
    LeftForward = ("Left Forward", "LF", "Attacker")
    Striker = ("Striker", "ST", "Attacker")
    RightForward = ("Right Forward", "RF", "Attacker")

    def __init__(self, long_name, code, parent):
        self.long_name = long_name
        self.code = code
        self._parent = parent

    @property
    def parent(self):
        if self._parent:
            return PositionType[self._parent]
        return None

    def is_subtype_of(self, other):
        current = self
        while current is not None:
            if current == other:
                return True
            current = current.parent
        return False

    def __str__(self):
        return self.long_name

    @classmethod
    def unknown(cls) -> "PositionType":
        return cls.Unknown



from typing import List, Optional, Union

from kloppy.config import get_config
from kloppy.domain import EventDataset, EventFactory, TrackingDataset
from kloppy.exceptions import KloppyError
from kloppy.infra.serializers.event.metrica import (
    MetricaJsonEventDataDeserializer,
    MetricaJsonEventDataInputs,
)
from kloppy.infra.serializers.tracking.metrica_csv import (
    MetricaCSVTrackingDataDeserializer,
    MetricaCSVTrackingDataInputs,
)
from kloppy.infra.serializers.tracking.metrica_epts import (
    MetricaEPTSTrackingDataDeserializer,
    MetricaEPTSTrackingDataInputs,
)
from kloppy.io import FileLike, open_as_file


def load_tracking_csv(
    home_data: FileLike,
    away_data: FileLike,
    sample_rate: Optional[float] = None,
    limit: Optional[int] = None,
    coordinates: Optional[str] = None,
) -> TrackingDataset:
    """
    Load Metrica Sports CSV tracking data.

    Args:
        home_data: The raw tracking data for the home team.
        away_data: The raw tracking data for the away team.
        sample_rate: Sample the data at a specific rate.
        limit: Limit the number of frames to load to the first `limit` frames.
        coordinates: The coordinate system to use.

    Returns:
        The parsed tracking data.
    """
    deserializer = MetricaCSVTrackingDataDeserializer(
        sample_rate=sample_rate, limit=limit, coordinate_system=coordinates
    )
    with open_as_file(home_data) as home_data_fp, open_as_file(
        away_data
    ) as away_data_fp:
        return deserializer.deserialize(
            inputs=MetricaCSVTrackingDataInputs(
                home_data=home_data_fp, away_data=away_data_fp
            )
        )


def load_tracking_epts(
    meta_data: FileLike,
    raw_data: FileLike,
    sample_rate: Optional[float] = None,
    limit: Optional[int] = None,
    coordinates: Optional[str] = None,
) -> TrackingDataset:
    """
    Load Metrica Sports EPTS tracking data.

    Args:
        meta_data: A xml feed containing the match meta data.
        raw_data: A feed containing the raw tracking data in the EPTS format.
        sample_rate: Sample the data at a specific rate.
        limit: Limit the number of frames to load to the first `limit` frames.
        coordinates: The coordinate system to use.

    Returns:
        The parsed tracking data.
    """
    deserializer = MetricaEPTSTrackingDataDeserializer(
        sample_rate=sample_rate, limit=limit, coordinate_system=coordinates
    )
    with open_as_file(raw_data) as raw_data_fp, open_as_file(
        meta_data
    ) as meta_data_fp:
        return deserializer.deserialize(
            inputs=MetricaEPTSTrackingDataInputs(
                raw_data=raw_data_fp, meta_data=meta_data_fp
            )
        )


def load_event(
    event_data: FileLike,
    meta_data: FileLike,
    event_types: Optional[List[str]] = None,
    coordinates: Optional[str] = None,
    event_factory: Optional[EventFactory] = None,
) -> EventDataset:
    """Load Metrica Sports JSON event data.

    Args:
        event_data: A json feed containing the raw event data.
        meta_data: A xml feed containing the match meta data.
        event_types: A list of event types to load.
        coordinates: The coordinate system to use.
        event_factory: A custom event factory.

    Returns:
        The parsed event data.
    """
    deserializer = MetricaJsonEventDataDeserializer(
        event_types=event_types,
        coordinate_system=coordinates,
        event_factory=event_factory or get_config("event_factory"),
    )

    with open_as_file(event_data) as event_data_fp, open_as_file(
        meta_data
    ) as meta_data_fp:
        return deserializer.deserialize(
            inputs=MetricaJsonEventDataInputs(
                event_data=event_data_fp, meta_data=meta_data_fp
            )
        )


def load_open_data(
    match_id: Union[str, int] = "1",
    sample_rate: Optional[float] = None,
    limit: Optional[int] = None,
    coordinates: Optional[str] = None,
) -> TrackingDataset:
    """Load Metrica Sports open data.

    This function loads tracking data directly from Metrica's open data
    GitHub repository.

    Args:
        match_id: The id of the match to load data for.
        sample_rate: Sample the data at a specific rate.
        limit: Limit the number of frames to load to the first `limit` frames.
        coordinates: The coordinate system to use.

    Returns:
        The parsed event data.
    """
    if match_id == "1" or match_id == 1:
        return load_tracking_csv(
            home_data="https://raw.githubusercontent.com/metrica-sports/sample-data/"
            "master/data/Sample_Game_1/Sample_Game_1_RawTrackingData_Home_Team.csv",
            away_data="https://raw.githubusercontent.com/metrica-sports/sample-data/"
            "master/data/Sample_Game_1/Sample_Game_1_RawTrackingData_Away_Team.csv",
            sample_rate=sample_rate,
            limit=limit,
            coordinates=coordinates,
        )
    elif match_id == "2" or match_id == 2:
        return load_tracking_csv(
            home_data="https://raw.githubusercontent.com/metrica-sports/sample-data/"
            "master/data/Sample_Game_2/Sample_Game_2_RawTrackingData_Home_Team.csv",
            away_data="https://raw.githubusercontent.com/metrica-sports/sample-data/"
            "master/data/Sample_Game_2/Sample_Game_2_RawTrackingData_Away_Team.csv",
            sample_rate=sample_rate,
            limit=limit,
            coordinates=coordinates,
        )
    elif match_id == "3" or match_id == 3:
        return load_tracking_epts(
            meta_data="https://raw.githubusercontent.com/metrica-sports/sample-data/"
            "master/data/Sample_Game_3/Sample_Game_3_metadata.xml",
            raw_data="https://raw.githubusercontent.com/metrica-sports/sample-data/"
            "master/data/Sample_Game_3/Sample_Game_3_tracking.txt",
            sample_rate=sample_rate,
            limit=limit,
            coordinates=coordinates,
        )
    else:
        raise KloppyError(
            f"Don't know where to fetch Metrica open data for {match_id}"
        )



from datetime import timedelta

from pandas import DataFrame
from pandas._testing import assert_frame_equal

from kloppy import sportscode
from kloppy.domain import Period
from kloppy.infra.serializers.code.sportscode import SportsCodeSerializer


class TestXMLCodeTracking:
    def test_correct_deserialization(self, base_dir):
        dataset = sportscode.load(base_dir / "files/code_xml.xml")

        assert len(dataset.metadata.periods) == 1

        assert dataset.metadata.periods[0].start_timestamp == timedelta(
            seconds=0
        )
        assert (
            dataset.metadata.periods[0].end_timestamp
            == dataset.codes[-1].end_timestamp
        )

        assert len(dataset.codes) == 3
        assert dataset.codes[0].code_id == "P1"
        assert dataset.codes[0].code == "PASS"
        assert dataset.codes[0].timestamp == timedelta(seconds=3.6)
        assert dataset.codes[0].end_timestamp == timedelta(seconds=9.7)
        assert dataset.codes[0].labels == {
            "Team": "Henkie",
            "Packing.Value": 1,
            "Receiver": "Klaas Nøme",
        }

        dataframe = dataset.to_df(engine="pandas")

        expected_data_frame = DataFrame.from_dict(
            {
                "code_id": ["P1", "P2", "P3"],
                "period_id": [1, 1, 1],
                "timestamp": [
                    timedelta(seconds=3.6),
                    timedelta(seconds=68.3),
                    timedelta(seconds=103.6),
                ],
                "end_timestamp": [
                    timedelta(seconds=9.7),
                    timedelta(seconds=74.5),
                    timedelta(seconds=109.6),
                ],
                "code": ["PASS", "PASS", "SHOT"],
                "Team": ["Henkie", "Henkie", "Henkie"],
                "Packing.Value": [1, 3, None],
                "Receiver": ["Klaas Nøme", "Piet", None],
                "Expected.Goal.Value": [None, None, 0.13],
            }
        )

        assert_frame_equal(dataframe, expected_data_frame)

    def test_correct_serialization(self, base_dir):
        dataset = sportscode.load(base_dir / "files/code_xml.xml")

        del dataset.codes[2:]

        # Make sure that data in Period 2 get the timestamp corrected
        dataset.metadata.periods = [
            Period(
                id=1,
                start_timestamp=timedelta(seconds=0),
                end_timestamp=timedelta(minutes=45),
            ),
            Period(
                id=2,
                start_timestamp=timedelta(minutes=45, seconds=10),
                end_timestamp=timedelta(minutes=90),
            ),
        ]
        dataset.codes[1].period = dataset.metadata.periods[1]

        serializer = SportsCodeSerializer()
        output = serializer.serialize(dataset)

        expected_output = """<?xml version='1.0' encoding='utf-8'?>
<file>
  <ALL_INSTANCES>
    <instance>
      <ID>P1</ID>
      <start>3.6</start>
      <end>9.7</end>
      <code>PASS</code>
      <label>
        <group>Team</group>
        <text>Henkie</text>
      </label>
      <label>
        <group>Packing.Value</group>
        <text>1</text>
      </label>
      <label>
        <group>Receiver</group>
        <text>Klaas Nøme</text>
      </label>
    </instance>
    <instance>
      <ID>P2</ID>
      <start>2768.3</start>
      <end>2774.5</end>
      <code>PASS</code>
      <label>
        <group>Team</group>
        <text>Henkie</text>
      </label>
      <label>
        <group>Packing.Value</group>
        <text>3</text>
      </label>
      <label>
        <group>Receiver</group>
        <text>Piet</text>
      </label>
    </instance>
  </ALL_INSTANCES>
</file>
"""
        expected_output = bytes(expected_output, "utf-8")
        assert output == expected_output
